# -*- coding: utf-8 -*-
import optparse
import logging

from pyramid_oereb.contrib.data_sources.create_tables import \
    create_tables_from_standard_configuration


logging.basicConfig()
log = logging.getLogger(__name__)


def create_standard_tables():
    parser = optparse.OptionParser(
        usage='usage: %prog [options]',
        description='Create all content for the standard database'
    )
    parser.add_option(
        '-c', '--configuration',
        dest='configuration',
        metavar='YAML',
        type='string',
        help='The absolute path to the configuration yaml file.'
    )
    parser.add_option(
        '-s', '--section',
        dest='section',
        metavar='SECTION',
        type='string',
        default='pyramid_oereb',
        help='The section which contains configuration (default is: pyramid_oereb).'
    )
    parser.add_option(
        '-T', '--tables-only',
        dest='tables_only',
        action='store_true',
        default=False,
        help='Use this flag to skip the creation of the schema.'
    )
    parser.add_option(
        '--sql-file',
        type='string',
        help='Generate an SQL file.'
    )
    parser.add_option(
        '--c2ctemplate-style',
        dest='c2ctemplate_style',
        action='store_true',
        default=False,
        help='Is the yaml file using a c2ctemplate style (starting with vars)'
    )
    parser.add_option(
        '-w', '--over-write',
        dest='append_to_sql',
        action='store_true',
        default=False,
        help='Setting this will overwrite the given sql-file (defaults to append to the file).'
    )

    options, args = parser.parse_args()
    if not options.configuration:
        parser.error('No configuration file set.')

    config_source = 'pyramid_oereb.contrib.data_sources.standard.sources.plr.DatabaseSource'
    if options.sql_file is None:
        create_tables_from_standard_configuration(
            configuration_yaml_path=options.configuration,
            source_class=config_source,
            section=options.section,
            c2ctemplate_style=options.c2ctemplate_style,
            tables_only=options.tables_only
        )
    else:
        append_to_sql = 'w' if options.append_to_sql else 'a'
        with open(options.sql_file, append_to_sql) as sql_file:
            create_tables_from_standard_configuration(
                configuration_yaml_path=options.configuration,
                source_class=config_source,
                section=options.section,
                c2ctemplate_style=options.c2ctemplate_style,
                sql_file=sql_file
            )



# -*- coding: utf-8 -*-
import logging

from operator import attrgetter

from pyramid.path import DottedNameResolver

from pyramid_oereb.core.config import Config
from pyramid_oereb.core.records.plr import PlrRecord
from pyramid_oereb.core.readers.extract import ExtractReader
from pyramid_oereb.core.readers.real_estate import RealEstateReader


log = logging.getLogger(__name__)


class Processor(object):

    def __init__(self, real_estate_reader, plr_sources, extract_reader):
        """
        The Processor class is directly bound to the get_extract_by_id service in this application. It's task
        is to unsnarl the difficult model of the oereb extract and handle all objects inside this extract
        correctly. In addition it provides an easy to use method interface to access the information.
        It is also used to wrap all accessors in one point to have a processing interface.

        Args:
            real_estate_reader (pyramid_oereb.lib.readers.real_estate.RealEstateReader): The
                real estate reader instance for runtime use.
            plr_sources (list of pyramid_oereb.standard.sources.plr.DatabaseSource): The
                public law restriction source instances for runtime use wrapped in a list.
            extract_reader (pyramid_oereb.lib.readers.extract.ExtractReader): The extract reader
                instance for runtime use.
        """
        self._real_estate_reader_ = real_estate_reader
        self._plr_sources_ = plr_sources
        self._extract_reader_ = extract_reader

    def filter_published_documents(self, record):
        """
        Filter only published documents.

        Args:
            record (pyramid_oereb.core.records.plr.PlrRecord or
                pyramid_oereb.core.records.documents.DocumentRecord): The public law restriction or
                document record.
        """
        published_docs = list()
        if isinstance(record, PlrRecord):
            for doc in record.documents:
                if doc.published:
                    published_docs.append(doc)
                else:
                    log.debug("filtering out non-published document {}".format(doc))
            record.documents = published_docs
        return record

    def filter_documents_by_fosnr(self, record, fosnr):
        """
        Filter documents that are only relevant for one municipality.

        Args:
            record (pyramid_oereb.core.records.plr.PlrRecord or
                pyramid_oereb.core.records.documents.DocumentRecord): The public law restriction or
                document record.
            fosnr (int): The fosnr (= id bfs) of the municipality.
        """
        relevant_docs = list()
        if isinstance(record, PlrRecord):
            for doc in record.documents:
                if doc.only_in_municipality in [None, fosnr]:
                    relevant_docs.append(doc)
                else:
                    log.debug("filtering out document (not relevant for this municipality) {}".format(doc))
            record.documents = relevant_docs
        return record

    def plr_tolerance_check(self, extract):
        """
        The function checks if the found plr results exceed the minimal surface or length
        value defined in the configuration and therefore should be represented in the extract
        or if they should be considered as 'false trues' and be removed from the results.
        In addition, this function filters these documents out that are not published or that are not
        relevant for the municipality the real estate is located in.

        Args:
            extract (pyramid_oereb.lib.records.extract.ExtractRecord): The extract in it's
                unvalidated form

        Returns:
            pyramid_oereb.lib.records.extract.ExtractRecord: Returns the updated extract
        """

        real_estate = extract.real_estate
        inside_plrs = []
        outside_plrs = []

        for public_law_restriction in real_estate.public_law_restrictions:
            if isinstance(public_law_restriction, PlrRecord) and public_law_restriction.published:
                # Test if the geometries list is now empty - if so remove plr from plr list
                if public_law_restriction.calculate(real_estate, Config.get('geometry_types')):
                    log.debug("plr_tolerance_check: keeping as potentially concerned plr {}".
                              format(public_law_restriction))
                    public_law_restriction = self.filter_documents_by_fosnr(public_law_restriction,
                                                                            real_estate.fosnr)
                    public_law_restriction = self.filter_published_documents(public_law_restriction)
                    inside_plrs.append(public_law_restriction)
                else:
                    log.debug("plr_tolerance_check: removing from the concerned plrs {}".
                              format(public_law_restriction))
                    outside_plrs.append(public_law_restriction)

        # Check if theme is concerned
        def is_inside_plr(theme_code):
            for plr in inside_plrs:
                if plr.theme.code == theme_code:
                    return True
            return False

        # Ensure only ConcernedThemes are contained in PLRs
        themes_to_move = []
        for i, theme in enumerate(extract.concerned_theme):
            if not is_inside_plr(theme.code):
                themes_to_move.append(i)

        if len(themes_to_move) > 0:
            themes_to_move.reverse()
            for idx in themes_to_move:
                new_not_concerned_theme = extract.concerned_theme.pop(idx)
                log.debug("plr_tolerance_check() moving from concerned_theme to not_concerned_theme: {}"
                          .format(new_not_concerned_theme)
                          )
                extract.not_concerned_theme.append(new_not_concerned_theme)
            # Need to reorder, because order must stay exactly as defined in configuration
            extract.not_concerned_theme = sorted(extract.not_concerned_theme, key=attrgetter('extract_index'))

        real_estate.public_law_restrictions = self.get_legend_entries(inside_plrs, outside_plrs)
        return extract

    @staticmethod
    def view_service_handling(real_estate, images, extract_format, language):
        """
        Handles all view service related stuff. In the moment this is:
            * construction of the correct url (reference_wms, multilingual) depending on the real estate
            * downloading of the image (if parameter was set) for the requested or default language

        Args:
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord):
                The real estate record to be updated.
            images (bool): Switch whether the images should be downloaded or not.
            extract_format (string): The format currently used. For 'pdf' format,
                the used map size will be adapted to the pdf format.
            language (string or None): Which language of the reference WMS should be used

        Returns:
            pyramid_oereb.lib.records.real_estate.RealEstateRecord: The updated extract.
        """
        language = language or Config.get('default_language')
        map_size = Config.get_map_size(extract_format)
        bbox = Config.get_bbox(real_estate.limit)
        real_estate.plan_for_land_register.get_full_wms_url(
            language,
            map_size[0],
            map_size[1],
            bbox
        )
        real_estate.plan_for_land_register_main_page.get_full_wms_url(
            language,
            map_size[0],
            map_size[1],
            bbox
        )
        if images:
            real_estate.plan_for_land_register.download_wms_content(language)
            real_estate.plan_for_land_register_main_page.download_wms_content(language)

        for public_law_restriction in real_estate.public_law_restrictions:
            public_law_restriction.view_service.get_full_wms_url(language, map_size[0], map_size[1], bbox)
            if images:
                public_law_restriction.view_service.download_wms_content(language)
        return real_estate

    @staticmethod
    def get_legend_entries(inside_plrs, outside_plrs):
        """
        We need to apply the right legend entries to each plr record which is intersecting the real estate.
        The result will be the "other legend".
        Since we already have a list of legend entries which are in the view bbox from calculated in the
        plr source, its only necessary to omit the legend entries which are on the real estate from the list
        of "other legend".

        Args:
            inside_plrs (list of pyramid_oereb.lib.records.plr.PlrRecord): The PLR's which are intersecting
                the real estate
            outside_plrs (list of pyramid_oereb.lib.records.plr.PlrRecord): The PLR's which are in the BBOX
                of the view but not intersecting the real estate
        Returns:
            list of pyramid_oereb.core.records.plr.PlrRecord: The updated records with the hopefully correct
                legend entries assigned.
        """
        # first we create a dict of theme codes which holds all the type_codes we will remove later from
        # the legend entries.
        type_codes_to_remove = {}
        for inside_plr in inside_plrs:
            view_service_id = str(inside_plr.view_service_id)
            theme_code = inside_plr.theme.code
            type_code = inside_plr.type_code
            if not type_codes_to_remove.get(theme_code):
                type_codes_to_remove[theme_code] = {}
            if not type_codes_to_remove[theme_code].get(view_service_id):
                type_codes_to_remove[theme_code][view_service_id] = {
                    'codes_to_rm': [],
                    'legend_entries': None
                }

            if type_code not in type_codes_to_remove[theme_code][view_service_id]['codes_to_rm']:
                type_codes_to_remove[theme_code][view_service_id]['codes_to_rm'].append(type_code)
            if not type_codes_to_remove[theme_code][view_service_id]['legend_entries']:
                type_codes_to_remove[theme_code][view_service_id]['legend_entries'] = inside_plr.\
                    view_service.legends

        for key in type_codes_to_remove.keys():
            for view_service_key in type_codes_to_remove[key]:
                for legend_entry in list(type_codes_to_remove[key][view_service_key]['legend_entries']):
                    if legend_entry.type_code in type_codes_to_remove[key][view_service_key]['codes_to_rm']:
                        type_codes_to_remove[key][view_service_key]['legend_entries'].remove(legend_entry)

        for inside_plr in inside_plrs:
            theme_code = inside_plr.theme.code
            view_service_id = str(inside_plr.view_service_id)
            inside_plr.view_service.legends = type_codes_to_remove[
                theme_code
            ][view_service_id]['legend_entries']

        return inside_plrs

    @property
    def real_estate_reader(self):
        """
        Returns:
            pyramid_oereb.lib.readers.real_estate.RealEstateReader: The real estate reader
            instance.
        """
        return self._real_estate_reader_

    @property
    def plr_sources(self):
        """
        Returns:
            list of pyramid_oereb.core.sources.plr.DatabaseSource: The list of plr
            source instances.
        """
        return self._plr_sources_

    @property
    def extract_reader(self):
        """
        Returns:
            pyramid_oereb.lib.readers.extract.ExtractReader: The extract reader instance.
        """
        return self._extract_reader_

    def process(self, real_estate, params, sld_url):
        """
        Central processing method to hook in from webservice.

        Args:
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real
                estate reader to obtain the real estates record.
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract
                request.
            sld_url (str): The URL which provides the sld to style and filter the highlight of the real
                estate.

        Returns:
            pyramid_oereb.lib.records.extract.ExtractRecord: The generated extract record.
        """
        log.debug("process() start")
        municipality = Config.municipality_by_fosnr(real_estate.fosnr)
        extract_raw = self._extract_reader_.read(params, real_estate, municipality)
        extract = self.plr_tolerance_check(extract_raw)

        resolver = DottedNameResolver()
        sort_within_themes_method_string = Config.get('extract').get('sort_within_themes_method')
        if sort_within_themes_method_string:
            sort_within_themes_method = resolver.resolve(sort_within_themes_method_string)
            extract = sort_within_themes_method(extract)
        else:
            log.info("No configuration is provided for extract sort_within_themes_method;"
                     " no further sorting is applied.")

        # the selection of view services is done after the tolerance check. This enables us to take
        # care about the circumstance that after tolerance check plrs will be dismissed which were
        # recognized as intersecting before. To avoid this the tolerance check is gathering all plrs
        # intersecting and not intersecting and starts the legend entry sorting after.
        self.view_service_handling(extract.real_estate, params.images, params.format, params.language)

        extract.disclaimers = Config.disclaimers
        extract.glossaries = Config.glossaries
        log.debug("process() done, returning extract.")
        return extract


def create_processor():
    """
    Creates and returns a processor based on the application configuration.
    You should use one (and only one) processor per request. Otherwise some results can be mixed or
    missing.

    Returns:
        pyramid_oereb.lib.processor.Processor: A processor.
    """

    real_estate_config = Config.get_real_estate_config()

    plr_cadastre_authority = Config.get_plr_cadastre_authority()

    real_estate_reader = RealEstateReader(
        real_estate_config.get('source').get('class'),
        **real_estate_config.get('source').get('params')
    )

    plr_sources = []
    for plr in Config.get('plrs'):
        plr_source_class = DottedNameResolver().maybe_resolve(plr.get('source').get('class'))
        plr_sources.append(plr_source_class(**plr))

    extract_reader = ExtractReader(
        plr_sources,
        plr_cadastre_authority
    )

    return Processor(
        real_estate_reader=real_estate_reader,
        plr_sources=plr_sources,
        extract_reader=extract_reader,
    )



# -*- coding: utf-8 -*-
"""
This Package provides all models fitting to the standard database configuration. It is separated by the 17
mandatory topics the federation asks for.
"""

NAMING_CONVENTION = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
}



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.glossary import GlossaryRecord


class GlossaryBaseSource(Base):
    """
    Base class for address sources.

    Attributes:
        records (list of pyramid_oereb.lib.records.glossary.GlossaryRecord): List of glossary records.
    """
    _record_class_ = GlossaryRecord

    def read(self):
        """
        Every glossary source has to implement a read method. This method must accept no parameters. Because
        it should deliver all items available.
        If you want adapt to your own source for glossaries, this is the point where to hook in.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
import warnings

from pyramid_oereb.core import b64

from pyramid_oereb.core.records.image import ImageRecord


class LogoRecord(object):
    """
    Represents a logo with its code and its base64 encoded string (multilingual).

    Attributes:
        code (str of unicode): The code for the logo.
        image_dict (dict of pyramid_oereb.lib.records.image.ImageRecord): The image encoded
            as base64 (multilingual).
    """
    def __init__(self, code, image_dict):
        """
        Args:
            code (str of unicode): The code for the logo.
            image_dict (dict of pyramid_oereb.lib.records.image.ImageRecord): The image encoded
                as base64 (multilingual).
        """

        if not isinstance(code, str):
            warnings.warn('Type of "code" should be "str"')
        if not isinstance(image_dict, dict):
            warnings.warn('Type of "image_dict" should be "dict"')

        self.code = code
        self.image_dict = {}
        for key in image_dict.keys():
            self.image_dict[key] = ImageRecord(b64.decode(image_dict[key]))



# -*- coding: utf-8 -*-
import logging
import traceback

from pyramid.httpexceptions import HTTPInternalServerError

log = logging.getLogger(__name__)


def get_multilingual_element(value, language, not_null=True):
    """
        Method that takes a string or dict in order to retrieve the text
        value of the requested language.

        Args:
            values (str or dict): The multilingual values encoded as JSON.
            language (str): The language in which the multilingual element
                is to be returned.

        Returns:
            String: The equivalent string in the requested language.

        Raises:
            HTTPInternalServerError: If default language of requested value
                is not available.
        """
    multilingual_value = value.get(language, None)

    if multilingual_value is None and not_null:
        msg = 'Default language "{language}" is not available in: \n {element}'\
            .format(language=language, element=value)
        log.error(msg)
        log.info("Call stack of failing get_multilingual_element:")
        for line in traceback.format_stack():
            log.info("{}".format(line))

        raise HTTPInternalServerError()

    return multilingual_value




from pyramid_oereb.core.sources import BaseDatabaseSource
from pyramid_oereb.core.sources.data_integration import DataIntegrationRecord


class DatabaseSource(BaseDatabaseSource, DataIntegrationRecord):

    def read(self):
        """
        The read method to access the standard database structure. It uses SQL-Alchemy for querying. It does
        not accept any parameters nor it applies any filter on the database query. It simply loads all
        content from the configured model.
        """
        session = self._adapter_.get_session(self._key_)
        try:
            results = session.query(self._model_).all()

            self.records = list()
            for result in results:
                self.records.append(self._record_class_(
                    result.date,
                    result.checksum,
                    result.available
                ))
        finally:
            session.close()



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.map_layering import MapLayeringRecord


class MapLayeringBaseSource(Base):
    """
    Base class for address sources.

    Attributes:
        records (list of pyramid_oereb.lib.records.map_layering.MapLayeringRecord):
            List of map layering records.
    """
    _record_class_ = MapLayeringRecord

    def read(self):
        """
        Every map layering source has to implement a read method. This method must accept no parameters.
        Because it should deliver all items available.
        If you want adapt to your own source for map layering, this is the point where to hook in.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
import logging

from pyramid_oereb import Config
from pyramid_oereb.contrib.data_sources.oereblex.sources.document import OEREBlexSource
from pyramid_oereb.contrib.data_sources.standard.sources.plr import DatabaseSource
from sqlalchemy.orm import selectinload

log = logging.getLogger(__name__)


class DatabaseOEREBlexSource(DatabaseSource):
    """
    A source to get models documents attached to public law restrictions in replacement
    of standards documents. Be sure to use a model with an OEREBlex "geolink" integer
    column for plrs that use this source.
    """
    def __init__(self, **kwargs):
        """
        Keyword Arguments:
            name (str): The name. You are free to choose one.
            code (str): The official code. Regarding to the federal specifications.
            geometry_type (str): The geometry type. Possible are: POINT, POLYGON, LINESTRING,
                GEOMETRYCOLLECTION
            thresholds (dict): The configuration of limits and units used for processing.
            text (dict of str): The speaking title. It must be a dictionary containing language (as
                configured) as key and text as value.
            language (str): The language this public law restriction is originally shipped with.
            federal (bool): Switch if it is a federal topic. This will be taken into account in processing
                steps.
            source (dict): The configuration dictionary of the public law restriction
            hooks (dict of str): The hook methods: get_symbol, get_symbol_ref. They have to be provided as
                dotted string for further use with dotted name resolver of pyramid package.
            law_status (dict of str): The multiple match configuration to provide more flexible use of the
                federal specified classifiers "inKraft" or "AenderungMitVorwirkung"
                or "AenderungOhneVorwirkung".
        """
        super(DatabaseOEREBlexSource, self).__init__(**kwargs)
        config = Config.get_oereblex_config()
        config["code"] = self._plr_info.get('code')
        self._oereblex_source = OEREBlexSource(**config)
        self._queried_geolinks = {}

    @staticmethod
    def get_config_value_for_plr_code(url_param_config, plr_code):
        """
        Returns the appropriate configuration entry for a plr within a url_param_config section.

        Args:
            url_param_config (list of code and url_param): url parameters to use, per plr code
            plr_code (str): the plr code
        """
        for url_param_entry in url_param_config:
            if url_param_entry['code'] == plr_code:
                if 'url_param' in url_param_entry:
                    return url_param_entry['url_param']
                else:
                    log.warning("Incorrect configuration: missing url_param for entry {}".format(plr_code))
                    return None
        return None

    def get_document_records(self, params, public_law_restriction_from_db):
        """
        Override the parent's get_document_records method to obtain the models document instead.

        Returns:
            list of pyramid_oereb.core.records.documents.DocumentRecord: The documents created from
                the parsed OEREBlex response.
        """
        oereblex_params = None
        url_param_config = self._oereblex_source._url_param_config
        plr_code = self._plr_info.get('code')
        if url_param_config:
            oereblex_params = DatabaseOEREBlexSource.get_config_value_for_plr_code(url_param_config, plr_code)
        law_status = Config.get_law_status_by_data_code(
            plr_code,
            public_law_restriction_from_db.law_status
        )
        return self.document_records_from_oereblex(params, public_law_restriction_from_db.geolink,
                                                   law_status, oereblex_params)

    def document_records_from_oereblex(self, params, geolink, law_status, oereblex_params):
        """
        Create document records parsed from the OEREBlex response with the specified geoLink ID and appends
        them to the current public law restriction.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            geolink (int): The ID of the GEO-Link to request the documents for.
            law_status (pyramid_oereb.core.records.lawstatus.LawStatusRecord): The restriction's law status.
            oereblex_params (string): URL parameter to add to the models request

        Returns:
            list of pyramid_oereb.core.records.documents.DocumentRecord:
                The documents created from the parsed OEREBlex response.
        """
        log.debug("document_records_from_oereblex() start, GEO-Link {}, law status {}, oereblex_params {}"
                  .format(geolink, law_status.code, oereblex_params))
        identifier = '{}{}{}'.format(geolink, law_status.code, params.language)
        if identifier in self._queried_geolinks:
            log.debug('skip querying this geolink "{}" because it was fetched already.'.format(identifier))
            log.debug('use already queried instead')
        else:
            self._oereblex_source.read(params, geolink, law_status, oereblex_params)
            log.debug("document_records_from_oereblex() returning {} records"
                      .format(len(self._oereblex_source.records)))
            self._queried_geolinks[identifier] = self._oereblex_source.records
        return self._queried_geolinks[identifier]

    def collect_related_geometries_by_real_estate(self, session, real_estate):
        """
        Extracts all geometries in the topic which have spatial relation with the passed real estate

        Args:
            session (sqlalchemy.orm.Session): The requested clean session instance ready for use
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real
                estate in its record representation.

        Returns:
            list: The result of the related geometries unique by the public law restriction id
        """
        return self.handle_collection(session, real_estate.limit).distinct(
            self._model_.public_law_restriction_id
        ).options(
            selectinload(self.models.Geometry.public_law_restriction)
            .selectinload(self.models.PublicLawRestriction.geometries),
            selectinload(self.models.Geometry.public_law_restriction)
            .selectinload(self.models.PublicLawRestriction.legend_entry),
            selectinload(self.models.Geometry.public_law_restriction)
            .selectinload(self.models.PublicLawRestriction.view_service),
            selectinload(self.models.Geometry.public_law_restriction)
            .selectinload(self.models.PublicLawRestriction.responsible_office),
        ).all()



# -*- coding: utf-8 -*-
import logging
import importlib
import binascii

from geoalchemy2.shape import to_shape, from_shape
from shapely.geometry import Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon, \
    GeometryCollection
from sqlalchemy import or_
from sqlalchemy.orm import selectinload

from pyramid_oereb import Config
from pyramid_oereb.core import b64
from pyramid_oereb.core.records.image import ImageRecord
from pyramid_oereb.core.records.plr import EmptyPlrRecord
from pyramid_oereb.core.sources import BaseDatabaseSource
from pyramid_oereb.core.sources.plr import PlrBaseSource
from pyramid_oereb.contrib.data_sources.interlis_2_3.interlis_2_3_utils import from_multilingual_text_to_dict
from pyramid_oereb.contrib.data_sources.interlis_2_3.interlis_2_3_utils import from_multilingual_uri_to_dict
from pyramid_oereb.contrib import eliminate_duplicated_document_records

log = logging.getLogger(__name__)


class StandardThemeConfigParser(object):

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def model_factory_path(self):
        return self.kwargs.get('source').get('params').get('model_factory')

    @property
    def schema_name(self):
        return self.kwargs.get('source').get('params').get('schema_name')

    @property
    def geometry_type(self):
        return self.kwargs.get('geometry_type')

    @property
    def srid(self):
        return Config.get_srid()

    @property
    def db_connection(self):
        return self.kwargs.get('source').get('params').get('db_connection')

    def get_model_factory(self):
        module_elements = Config.extract_module_function(
            self.model_factory_path
        )
        return getattr(
            importlib.import_module(module_elements['module_path']),
            module_elements['function_name']
        )

    def get_models(self):
        model_factory = self.get_model_factory()
        return model_factory(
            self.schema_name,
            self.geometry_type,
            self.srid,
            self.db_connection
        )


def parse_multiple_standard_themes(config):
    """
    Args:
        config (pyramid_oereb.standard.config.Config): The config object of the application.

    Returns:
        dict containing the factory_model for every plr code.
    """
    plrs = config._config.get('plrs')
    themes = {}
    for plr in plrs:
        config_parser = StandardThemeConfigParser(**plr)
        themes[plr['code']] = config_parser.get_models()
    return themes


class DatabaseSource(BaseDatabaseSource, PlrBaseSource):
    def __init__(self, **kwargs):
        """
        Keyword Arguments:
            name (str): The name. You are free to choose one.
            code (str): The official code. Regarding to the federal specifications.
            geometry_type (str): The geometry type. Possible are: POINT, POLYGON, LINESTRING,
                GEOMETRYCOLLECTION
            thresholds (dict): The configuration of limits and units used for processing.
            text (dict of str): The speaking title. It must be a dictionary containing language (as
                configured) as key and text as value.
            language (str): The language this public law restriction is originally shipped with.
            federal (bool): Switch if it is a federal topic. This will be taken into account in processing
                steps.
            source (dict): The configuration dictionary of the public law restriction
            hooks (dict of str): The hook methods: get_symbol, get_symbol_ref. They have to be provided as
                dotted string for further use with dotted name resolver of pyramid package.
            law_status (dict of str): The configuration dictionary of the law status. It consists of
                the code and text which must be a dictionary containing language (as configured)
                as key and text as value.
        """
        config_parser = StandardThemeConfigParser(**kwargs)
        self.models = config_parser.get_models()
        bds_kwargs = {
            'model': self.models.Geometry,
            'db_connection': kwargs.get('source').get('params').get('db_connection')
        }

        BaseDatabaseSource.__init__(self, **bds_kwargs)
        PlrBaseSource.__init__(self, **kwargs)

        self.legend_entry_model = self.models.LegendEntry
        self.datasource = []

        self._tolerance = self._plr_info.get('tolerance')

    def from_db_to_legend_entry_record(self, legend_entry_from_db):
        theme = Config.get_theme_by_code_sub_code(legend_entry_from_db.theme, legend_entry_from_db.sub_theme)
        legend_entry_record = self._legend_entry_record_class(
            ImageRecord(
                b64.decode(
                    binascii.b2a_base64(legend_entry_from_db.symbol).decode('ascii')
                )
            ),
            from_multilingual_text_to_dict(
                de=legend_entry_from_db.legend_text_de,
                fr=legend_entry_from_db.legend_text_fr,
                it=legend_entry_from_db.legend_text_it,
                rm=legend_entry_from_db.legend_text_rm,
                en=legend_entry_from_db.legend_text_en
            ),
            legend_entry_from_db.type_code,
            legend_entry_from_db.type_code_list,
            theme,
            view_service_id=legend_entry_from_db.view_service_id,
            identifier=legend_entry_from_db.t_id
        )
        return legend_entry_record

    def from_db_to_legend_entry_records(self, legend_entries_from_db, sub_theme=None):
        legend_entry_records = []
        if sub_theme is not None:
            for legend_entry_from_db in legend_entries_from_db:
                if legend_entry_from_db.sub_theme == sub_theme:
                    self.from_db_to_legend_entry_record(legend_entry_from_db)
        else:
            for legend_entry_from_db in legend_entries_from_db:
                legend_entry_records.append(
                    self.from_db_to_legend_entry_record(legend_entry_from_db)
                )

        return legend_entry_records

    def from_db_to_view_service_record(self, view_service_from_db, legend_entry_records, theme):
        multilingual_uri = from_multilingual_uri_to_dict(view_service_from_db.multilingual_uri)
        layer_index, layer_opacity = Config.get_index_and_opacity_of_view_service(multilingual_uri)
        view_service_record = self._view_service_record_class(
            multilingual_uri,
            layer_index,
            layer_opacity,
            Config.get('default_language'),
            Config.get('srid'),
            Config.get('proxies'),
            legends=legend_entry_records
        )
        return view_service_record

    def unwrap_multi_geometry_(self, law_status, published_from, published_until,
                               multi_geom, geo_metadata):
        unwrapped = []
        for geom in multi_geom.geoms:
            unwrapped.append(
                self._geometry_record_class(
                    law_status,
                    published_from,
                    published_until,
                    geom,
                    geo_metadata
                )
            )
        return unwrapped

    def unwrap_geometry_collection_(self, law_status, published_from, published_until,
                                    collection, geo_metadata):
        unwrapped = []
        if len(collection.geoms) > 1:
            raise AttributeError(u'There was more than one element in the GeometryCollection. '
                                 u'This is not supported!')
        for geom in collection.geoms:
            unwrapped.extend(
                self.create_geometry_records_(
                    law_status,
                    published_from,
                    published_until,
                    geom,
                    geo_metadata
                )
            )
        return unwrapped

    def create_geometry_records_(self, law_status, published_from, published_until,
                                 geom, geo_metadata):
        geometry_records = []

        # Process single geometries
        if isinstance(geom, (Point, LineString, Polygon)):
            geometry_records.append(
                self._geometry_record_class(
                    law_status,
                    published_from,
                    published_until,
                    geom,
                    geo_metadata
                )
            )

        # Process multi geometries
        elif isinstance(geom, (MultiPoint, MultiLineString, MultiPolygon)):
            geometry_records.extend(self.unwrap_multi_geometry_(
                law_status,
                published_from,
                published_until,
                geom,
                geo_metadata
            ))

        # Process geometry collections
        elif isinstance(geom, GeometryCollection):
            geometry_records.extend(self.unwrap_geometry_collection_(
                law_status,
                published_from,
                published_until,
                geom,
                geo_metadata
            ))

        else:
            log.warning('Received geometry with unsupported type: {0}. Skipped geometry.'.format(
                geom.type
            ))

        return geometry_records

    def from_db_to_geometry_records(self, geometries_from_db):
        geometry_records = []
        for geometry_from_db in geometries_from_db:
            # Create law status record
            law_status = Config.get_law_status_by_data_code(
                    self._plr_info.get('code'),
                    geometry_from_db.law_status
                )

            if geometry_from_db.point is not None:
                geom = geometry_from_db.point
            elif geometry_from_db.line is not None:
                geom = geometry_from_db.line
            elif geometry_from_db.surface is not None:
                geom = geometry_from_db.surface

            # Create geometry records
            geometry_records.extend(self.create_geometry_records_(
                law_status,
                geometry_from_db.published_from,
                geometry_from_db.published_until,
                to_shape(geom),
                geometry_from_db.geo_metadata
            ))

        return geometry_records

    def from_db_to_office_record(self, offices_from_db):
        office_record = self._office_record_class(
            from_multilingual_text_to_dict(
                de=offices_from_db.name_de,
                fr=offices_from_db.name_fr,
                it=offices_from_db.name_it,
                rm=offices_from_db.name_rm,
                en=offices_from_db.name_en,
            ),
            offices_from_db.uid,
            from_multilingual_uri_to_dict(offices_from_db.multilingual_uri),
            offices_from_db.line1,
            offices_from_db.line2,
            offices_from_db.street,
            offices_from_db.number,
            offices_from_db.postal_code,
            offices_from_db.city
        )

        return office_record

    def from_db_to_document_records(self, documents_from_db):
        document_records = []
        for document in documents_from_db:
            office_record = self.from_db_to_office_record(document.responsible_office)
            law_status = Config.get_law_status_by_data_code(
                self._plr_info.get('code'),
                document.law_status
            )
            document_records.append(self._documents_record_class(
                document_type=Config.get_document_type_by_data_code(
                    self._plr_info.get('code'),
                    document.document_type
                ),
                index=document.index,
                law_status=law_status,
                title=from_multilingual_text_to_dict(
                        de=document.title_de,
                        fr=document.title_fr,
                        it=document.title_it,
                        rm=document.title_rm,
                        en=document.title_en),
                responsible_office=office_record,
                published_from=document.published_from,
                published_until=document.published_until,
                text_at_web=from_multilingual_uri_to_dict(document.multilingual_uri),
                abbreviation=from_multilingual_text_to_dict(
                        de=document.abbreviation_de,
                        fr=document.abbreviation_fr,
                        it=document.abbreviation_it,
                        rm=document.abbreviation_rm,
                        en=document.abbreviation_en),
                official_number=from_multilingual_text_to_dict(
                        de=document.official_number_de,
                        fr=document.official_number_fr,
                        it=document.official_number_it,
                        rm=document.official_number_rm,
                        en=document.official_number_en),
                only_in_municipality=document.only_in_municipality,
                article_numbers=None,
                file=None
            ))
        return document_records

    def from_db_to_plr_record(self, params, public_law_restriction_from_db, legend_entries_from_db):
        thresholds = self._plr_info.get('thresholds')
        min_length = thresholds.get('length').get('limit')
        length_unit = thresholds.get('length').get('unit')
        min_area = thresholds.get('area').get('limit')
        area_unit = thresholds.get('area').get('unit')
        legend_entry_record = self.from_db_to_legend_entry_record(
            public_law_restriction_from_db.legend_entry
        )
        legend_entry_records = self.from_db_to_legend_entry_records(
            legend_entries_from_db,
            legend_entry_record.theme.sub_code
        )
        symbol = ImageRecord(
            b64.decode(
                binascii.b2a_base64(public_law_restriction_from_db.legend_entry.symbol).decode('ascii')
            )
        )
        view_service_record = self.from_db_to_view_service_record(
            public_law_restriction_from_db.view_service,
            legend_entry_records,
            self._plr_info.get('code')
        )
        theme = Config.get_theme_by_code_sub_code(
            public_law_restriction_from_db.legend_entry.theme,
            public_law_restriction_from_db.legend_entry.sub_theme)
        document_records = eliminate_duplicated_document_records(
            theme.document_records,
            self.get_document_records(params, public_law_restriction_from_db)
        )
        geometry_records = self.from_db_to_geometry_records(public_law_restriction_from_db.geometries)
        law_status = Config.get_law_status_by_data_code(
            self._plr_info.get('code'),
            public_law_restriction_from_db.law_status
        )
        sub_theme_legend_entry_record = \
            legend_entry_record.theme if legend_entry_record.theme.sub_code else None
        plr_record = self._plr_record_class(
            Config.get_theme_by_code_sub_code(legend_entry_record.theme.code),
            legend_entry_record,
            law_status,
            public_law_restriction_from_db.published_from,
            public_law_restriction_from_db.published_until,
            self.from_db_to_office_record(public_law_restriction_from_db.responsible_office),
            symbol,
            view_service_record,
            geometry_records,
            sub_theme=sub_theme_legend_entry_record,
            type_code=public_law_restriction_from_db.legend_entry.type_code,
            type_code_list=public_law_restriction_from_db.legend_entry.type_code_list,
            documents=document_records,
            min_area=min_area,
            min_length=min_length,
            area_unit=area_unit,
            length_unit=length_unit,
            view_service_id=public_law_restriction_from_db.view_service.t_id,
            tolerance=self._tolerance
        )

        return plr_record

    def get_document_records(self, params, public_law_restriction_from_db):
        documents_from_db = []
        if not hasattr(public_law_restriction_from_db, 'legal_provisions'):
            raise AttributeError('The public_law_restriction implementation of type {} has no '
                                 'legal_provisions attribute. Check the model implementation.'
                                 .format(type(public_law_restriction_from_db)))
        for legal_provision in public_law_restriction_from_db.legal_provisions:
            documents_from_db.append(legal_provision.document)
        document_records = self.from_db_to_document_records(documents_from_db)
        return document_records

    def collect_related_geometries_by_real_estate(self, session, real_estate):
        """
        Extracts all geometries in the topic which have spatial relation with the passed real estate

        Args:
            session (sqlalchemy.orm.Session): The requested clean session instance ready for use
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real
                estate in its record representation.

        Returns:
            list: The result of the related geometries unique by the public law restriction id
        """
        if self._tolerance is None:
            query = session.query(self._model_).filter(
                or_(
                    self._model_.point.ST_Intersects(from_shape(real_estate.limit, srid=Config.get('srid'))),
                    self._model_.line.ST_Intersects(from_shape(real_estate.limit, srid=Config.get('srid'))),
                    self._model_.surface.ST_Intersects(from_shape(real_estate.limit, srid=Config.get('srid')))
                ))
        else:
            query = session.query(self._model_).filter(
                or_(
                    self._model_.point.ST_Distance(
                        from_shape(real_estate.limit, srid=Config.get('srid'))
                    ) < self._tolerance,
                    self._model_.line.ST_Distance(
                        from_shape(real_estate.limit, srid=Config.get('srid'))
                    ) < self._tolerance,
                    self._model_.surface.ST_Distance(
                        from_shape(real_estate.limit, srid=Config.get('srid'))
                    ) < self._tolerance
                ))
        return query.distinct(self._model_.public_law_restriction_id).options(
            selectinload(self.models.Geometry.public_law_restriction)
            .selectinload(self.models.PublicLawRestriction.geometries),
            selectinload(self.models.Geometry.public_law_restriction)
            .selectinload(self.models.PublicLawRestriction.legal_provisions)
            .selectinload(self.models.PublicLawRestrictionDocument.document)
            .selectinload(self.models.Document.multilingual_uri)
            .selectinload(self.models.MultilingualUri.localised_uri),
            selectinload(self.models.Geometry.public_law_restriction)
            .selectinload(self.models.PublicLawRestriction.legend_entry),
            selectinload(self.models.Geometry.public_law_restriction)
            .selectinload(self.models.PublicLawRestriction.view_service)
            .selectinload(self.models.ViewService.multilingual_uri)
            .selectinload(self.models.MultilingualUri.localised_uri),
            selectinload(self.models.Geometry.public_law_restriction)
            .selectinload(self.models.PublicLawRestriction.responsible_office)
            .selectinload(self.models.Office.multilingual_uri)
            .selectinload(self.models.MultilingualUri.localised_uri)
        ).all()

    def collect_legend_entries_by_bbox(self, session, bbox, law_status):
        """
        Extracts all legend entries in the topic which have spatial relation with the passed bounding box of
        visible extent.

        Args:
            session (sqlalchemy.orm.Session): The requested clean session instance ready for use
            bbox (shapely.geometry.base.BaseGeometry): The bbox to search the records.
            law_status (str): String of the law status for which the legend entries should be queried.

        Returns:
            list: The result of the related geometries unique by the public law restriction id and law status
        """
        distinct_legend_entry_ids = []
        geometries = session.query(self._model_).filter(
                      or_(
                       self._model_.point.ST_Intersects(from_shape(bbox, srid=Config.get('srid'))),
                       self._model_.line.ST_Intersects(from_shape(bbox, srid=Config.get('srid'))),
                       self._model_.surface.ST_Intersects(from_shape(bbox, srid=Config.get('srid')))
                      )).distinct(self._model_.public_law_restriction_id).options(
                        selectinload(self.models.Geometry.public_law_restriction)
                      ).all()

        for geometry in geometries:
            if geometry.public_law_restriction.legend_entry_id not in distinct_legend_entry_ids\
                    and geometry.public_law_restriction.law_status == law_status:
                distinct_legend_entry_ids.append(geometry.public_law_restriction.legend_entry_id)

        return session.query(self.legend_entry_model).filter(
            self.legend_entry_model.t_id.in_((distinct_legend_entry_ids))).all()

    def read(self, params, real_estate, bbox):
        """
        The read point which creates a extract, depending on a passed real estate.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real
                estate in its record representation.
            bbox (shapely.geometry.base.BaseGeometry): The bbox to search the records.
        """

        # Check if the plr is marked as available
        if Config.availability_by_theme_code_municipality_fosnr(self._plr_info['code'], real_estate.fosnr):
            session = self._adapter_.get_session(self._key_)
            try:
                if session.query(self._model_).count() == 0:
                    # We can stop here already because there are no items in the database
                    self.records = [EmptyPlrRecord(
                            Config.get_theme_by_code_sub_code(self._plr_info['code'])
                        )]
                else:
                    # We need to investigate more in detail

                    # Try to find geometries which have spatial relation with real estate
                    geometry_results = self.collect_related_geometries_by_real_estate(
                        session, real_estate
                    )
                    if len(geometry_results) == 0:
                        # We checked if there are spatially related elements in database. But there is none.
                        # So we can stop here.
                        self.records = [EmptyPlrRecord(
                            Config.get_theme_by_code_sub_code(self._plr_info['code'])
                        )]
                    else:
                        # We found spatially related elements. This means we need to extract the actual plr
                        # information related to the found geometries.
                        law_status_of_geometry = []
                        # get distinct values of law_status for all geometries found
                        for geometry in geometry_results:
                            if(geometry.public_law_restriction.law_status not in law_status_of_geometry):
                                law_status_of_geometry.append(geometry.public_law_restriction.law_status)

                        legend_entries_from_db = []
                        # get legend_entries per law_status
                        for law_status in law_status_of_geometry:
                            legend_entry_with_law_status = [
                                self.collect_legend_entries_by_bbox(session, bbox, law_status),
                                law_status
                            ]
                            legend_entries_from_db.append(legend_entry_with_law_status)

                        self.records = []
                        for geometry_result in geometry_results:
                            self.records.append(
                                self.from_db_to_plr_record(
                                    params,
                                    geometry_result.public_law_restriction,
                                    next(elem for elem in legend_entries_from_db
                                         if elem[1] == geometry_result.public_law_restriction.law_status)[0]
                                )
                            )

            finally:
                session.close()

        # Add empty record if topic is not available
        else:
            self.records = [EmptyPlrRecord(
                Config.get_theme_by_code_sub_code(self._plr_info['code']),
                has_data=False
            )]



# -*- coding: utf-8 -*-

import logging
# import yappi
import qrcode
import io
# import re

from pyramid.httpexceptions import HTTPBadRequest, HTTPFound, HTTPInternalServerError, HTTPNoContent, \
    HTTPNotFound
from pyramid.path import DottedNameResolver
from shapely.geometry import Point
from pyramid.renderers import render_to_response

from pyramid_oereb import route_prefix
from pyramid_oereb import Config
from pyreproj import Reprojector

from pyramid_oereb.core.processor import create_processor
from pyramid_oereb.core.readers.address import AddressReader
from pyramid_oereb.core.renderer import Base as Renderer
from timeit import default_timer as timer

from pyramid_oereb.contrib.stats.decorators import OerebStats

log = logging.getLogger(__name__)


class PlrWebservice(object):
    """
    This class provides the PLR webservice methods.

    Args:
        request (pyramid.request.Request or pyramid.testing.DummyRequest): The pyramid request instance.
    """

    _DEFAULT_FORMATS = ['xml', 'json']
    """list of str: The default formats for the service responses."""

    _EXTRACT_FORMATS = _DEFAULT_FORMATS + ['pdf', 'url']
    """list of str: The formats for the extract responses."""

    def __init__(self, request):
        self._request = request
        self._params = {k.upper(): v for k, v in request.params.items()}

    def get_versions(self):
        """
        Returns the available versions of this service.

        Returns:
            pyramid.response.Response: The `versions` response.
        """
        if route_prefix:
            endpoint = u'{application}/{route_prefix}'.format(
                application=self._request.application_url,
                route_prefix=route_prefix
            )
        else:
            endpoint = u'{application}'.format(application=self._request.application_url)
        versions = {
            u'GetVersionsResponse': {
                u'supportedVersion': [
                    {
                        u'version': u'2.0',
                        u'serviceEndpointBase': endpoint
                    }
                ]
            }
        }
        output_format = self.__validate_format_param__(self._DEFAULT_FORMATS)
        renderer_name = 'json' if output_format == 'json' else 'pyramid_oereb_versions_xml'
        response = render_to_response(renderer_name, versions, request=self._request)
        if output_format == 'json':
            response.content_type = 'application/json; charset=UTF-8'
        response.extras = OerebStats(service='GetVersions', output_format=output_format)
        return response

    def get_capabilities(self):
        """
        Returns the capabilities of this service.

        Returns:
            pyramid.response.Response: The `capabilities` response.
        """

        output_format = self.__validate_format_param__(self._DEFAULT_FORMATS)

        supported_languages = Config.get_language()
        themes = list()
        for theme in Config.get_themes():
            text = list()
            for lang in theme.title:
                if lang in supported_languages:
                    text.append({
                        'Language': lang,
                        'Text': theme.title[lang]
                    })
            themes.append({
                'Code': theme.code,
                'Text': text
            })
        capabilities = {
            u'GetCapabilitiesResponse': {
                u'topic': Config.get_themes() if output_format == 'xml' else themes,
                u'municipality': [record.fosnr for record in Config.municipalities],
                u'flavour': Config.get_flavour(),
                u'language': supported_languages,
                u'crs': [Config.get_crs()]
            }
        }

        renderer_name = 'json' if output_format == 'json' else 'pyramid_oereb_capabilities_xml'  # noqa: E501
        response = render_to_response(renderer_name, capabilities, request=self._request)
        if output_format == 'json':
            response.content_type = 'application/json; charset=UTF-8'
        response.extras = OerebStats(service='GetCapabilities', output_format=output_format)
        return response

    def get_egrid(self):
        """
        Validates the request parameters and calls the query method
        according to the specified parameters.

        Returns:
            pyramid.response.Response: The `getegrid` response.
        """
        try:
            output_format = self.__validate_format_param__(self._DEFAULT_FORMATS)
            service = None
            with_geometry = False
            if self.__has_params__(['GEOMETRY']):
                if self._params.get('GEOMETRY').lower() == 'true':
                    with_geometry = True
            params = Parameter(
                output_format,
                with_geometry=with_geometry
            )
            # Type A
            if self.__has_params__(['EN']):
                service = 'GetEgridCoord'
                records = self._get_egrid_coord(params)
            # Type B
            elif self.__has_params__(['IDENTDN', 'NUMBER']):
                service = 'GetEgridIdent'
                records = self._get_egrid_ident(params)
            # Type C
            elif self.__has_params__(['POSTALCODE', 'LOCALISATION', 'NUMBER']):
                service = 'GetEgridAddress'
                records = self._get_egrid_address(params)
            # Type D
            elif self.__has_params__(['GNSS']):
                service = 'GetEgridCoord'
                records = self._get_egrid_coord(params)
            # Raise exception
            else:
                raise HTTPBadRequest(
                    'Invalid parameters. You need one of the following combinations: '
                    'EN or GNSS or IDENTDN and NUMBER or POSTALCODE, LOCALISATION and NUMBER.'
                )
            response = self.__get_egrid_response__(records, params)
        except HTTPNoContent as err:
            response = HTTPNoContent('{}'.format(err))
        except HTTPBadRequest as err:
            response = HTTPBadRequest('{}'.format(err))
        response.extras = OerebStats(
            service=service,
            params=dict(self._params),
            output_format=output_format
        )
        return response

    def _get_egrid_coord(self, params):
        """
        Returns a list with the matched EGRIDs for the given coordinates.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameter object.

        Returns:
            list of pyramid_oereb.core.records.real_estate.RealEstateRecord:
                The list of all found records filtered by the passed criteria.
        """
        en = self._params.get('EN')
        gnss = self._params.get('GNSS')
        if en or gnss:
            geom_wkt = 'SRID={0};{1}'
            if en:
                geom_wkt = geom_wkt.format(
                    Config.get('srid'),
                    self.__parse_en__(en, buffer_dist=1.0).wkt
                )
            elif gnss:
                geom_wkt = geom_wkt.format(
                    Config.get('srid'),
                    self.__parse_gnss__(gnss).wkt
                )
            processor = create_processor()
            return processor.real_estate_reader.read(params, **{'geometry': geom_wkt})
        else:
            raise HTTPBadRequest('EN or GNSS must be defined.')

    def _get_egrid_ident(self, params):
        """
        Returns a list with the matched EGRIDs for the given NBIdent and property number.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameter object.

        Returns:
            list of pyramid_oereb.core.records.real_estate.RealEstateRecord:
                The list of all found records filtered by the passed criteria.
        """
        identdn = self._params.get('IDENTDN')
        number = self._params.get('NUMBER')
        if identdn and number:
            processor = create_processor()
            return processor.real_estate_reader.read(
                params,
                **{
                    'nb_ident': identdn,
                    'number': number
                }
            )
        else:
            raise HTTPBadRequest('IDENTDN and NUMBER must be defined.')

    def _get_egrid_address(self, params):
        """
        Returns a list with the matched EGRIDs for the given postal address.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameter object.

        Returns:
            list of pyramid_oereb.core.records.real_estate.RealEstateRecord:
                The list of all found records filtered by the passed criteria.
        """
        postalcode = self._params.get('POSTALCODE')
        localisation = self._params.get('LOCALISATION')
        number = self._params.get('NUMBER')
        if postalcode and localisation and number:
            reader = AddressReader(
                Config.get_address_config().get('source').get('class'),
                **Config.get_address_config().get('source').get('params')
            )
            addresses = reader.read(params, localisation, int(postalcode), number)
            if len(addresses) == 0:
                raise HTTPNoContent()
            geometry = 'SRID={srid};{wkt}'.format(
                srid=Config.get('srid'),
                wkt=addresses[0].geom.wkt
            )
            processor = create_processor()
            return processor.real_estate_reader.read(params, **{'geometry': geometry})
        else:
            raise HTTPBadRequest('POSTALCODE, LOCALISATION and NUMBER must be defined.')

    def get_extract_by_id(self):
        """
        Returns the extract in the specified format and flavour.

        Returns:
            pyramid.response.Response: The `extract` response.
        """
        start_time = timer()
        log.debug("get_extract_by_id() start")
        try:
            params = self.__validate_extract_params__()
            processor = create_processor()
            # read the real estate from configured source by the passed parameters
            real_estate_reader = processor.real_estate_reader
            if params.egrid:
                real_estate_records = real_estate_reader.read(params, egrid=params.egrid)
            elif params.identdn and params.number:
                real_estate_records = real_estate_reader.read(
                    params,
                    nb_ident=params.identdn,
                    number=params.number
                )
            else:
                raise HTTPBadRequest("Missing required argument")
            # check if result is strictly one (we queried with primary keys)
            if len(real_estate_records) == 1:

                # Redirect for format URL
                if params.format == 'url':
                    log.debug("get_extract_by_id() calling url")
                    return self.__redirect_to_dynamic_client__(real_estate_records[0])
                # yappi.set_clock_type("cpu")
                # yappi.start()
                extract = processor.process(
                    real_estate_records[0],
                    params,
                    self._request.route_url('{0}/sld'.format(route_prefix))
                )
                # yappi.get_func_stats().save('/workspace/processor_function_stats.prof', "pstat")
                # yappi.get_thread_stats().save('/workspace/processor_thread_stats.prof', "pstat")

                if params.format == 'json':
                    log.debug("get_extract_by_id() calling json")
                    response = render_to_response(
                        'pyramid_oereb_extract_json',
                        (extract, params),
                        request=self._request
                    )
                elif params.format == 'xml':
                    log.debug("get_extract_by_id() calling xml")
                    response = render_to_response(
                        'pyramid_oereb_extract_xml',
                        (extract, params),
                        request=self._request
                    )
                elif params.format == 'pdf':
                    log.debug("get_extract_by_id() calling pdf")
                    response = render_to_response(
                        'pyramid_oereb_extract_print',
                        (extract, params),
                        request=self._request
                    )
                else:
                    raise HTTPBadRequest("The format '{}' is wrong".format(params.format))
                end_time = timer()
                log.debug("DONE with extract, time spent: {} seconds".format(end_time - start_time))
            else:
                raise HTTPNoContent("No real estate found")
        except HTTPNoContent as err:
            response = HTTPNoContent('{}'.format(err))
        except HTTPBadRequest as err:
            response = HTTPBadRequest('{}'.format(err))
        try:
            response.extras = OerebStats(service='GetExtractById',
                                         output_format=params.format,
                                         params=vars(params))
        except UnboundLocalError:
            response.extras = OerebStats(service='GetExtractById', params={'error': response.message})
        except Exception:
            # if params is not set we get UnboundLocalError
            # or we could get ValueError
            # in any case, the logging should never crash the response delivery
            try:
                response.extras = OerebStats(service='GetExtractById', params={'error': response.message})
            except AttributeError:
                response.extras = OerebStats(service='GetExtractById')
        return response

    def __validate_extract_params__(self):
        """
        Validates the input parameters for get_extract_by_id.

        Returns:
            pyramid_oereb.views.webservice.Parameter: The validated parameters.
        """

        # Get and check format
        extract_format = self.__validate_format_param__(self._EXTRACT_FORMATS)

        # With geometry?
        with_geometry = False
        user_requested_geometry = False
        if self._params.get('GEOMETRY', 'false').lower() == 'true':
            with_geometry = True
            user_requested_geometry = True

        # Check for invalid combinations
        if extract_format == 'pdf' and user_requested_geometry:
            raise HTTPBadRequest('Geometry is not available for format PDF.')

        # If PDF is to be produced, check if geometry should be included
        # (this override can be needed for the print service. Note that, to be compliant to specification,
        # the URL for a PDF request should not contain the geometry parameter)
        if extract_format == 'pdf':
            with_geometry = Config.get('print', {}).get('with_geometry', True) or with_geometry

        # With images?
        with_images = self._params.get('WITHIMAGES', 'false').lower() == 'true'

        # Signed?
        signed = self._params.get('SIGNED', 'false').lower() == 'true'

        params = Parameter(
            extract_format,
            with_geometry=with_geometry,
            images=with_images,
            signed=signed,
            extract_url=self._request.url,
            qr_code_ref=self._request.route_url(
                '{0}/image/qrcode'.format(route_prefix),
                _query={"extract_url": self._request.url}
            )
        )

        # Get id
        if self.__has_params__(['EGRID']):
            params.set_egrid(self._params['EGRID'])
        elif self.__has_params__(['IDENTDN', 'NUMBER']):
            params.set_identdn(self._params['IDENTDN'])
            params.set_number(self._params['NUMBER'])
        else:
            raise HTTPBadRequest(
                'Invalid parameters. You need one of the following combinations: '
                'EGRID or IDENTDN and NUMBER.'
            )

        # Language
        if 'LANG' in self._params:
            language = str(self._params.get('LANG')).lower()
            if language not in Config.get_language():
                raise HTTPBadRequest(
                    'Requested language is not available. Following languages are '
                    'configured: {languages}. The '
                    'requested language was: {language}.'.format(
                        languages=str(Config.get_language()),
                        language=language
                    )
                )
            params.set_language(language)

        # Topics
        topics = self._params.get('TOPICS')
        if topics:
            params.set_topics(topics.split(','))

        return params

    def __validate_format_param__(self, accepted_formats):
        """
        Get format in the url and validate that it's one accepted.

        Args:
            accepted_formats (list): A list of accepted format (str).

        Returns:
            str: The validated format parameter.
        """
        output_format = self._request.matchdict.get('format', '').lower()
        if output_format not in accepted_formats:
            raise HTTPBadRequest('Invalid format: {0}'.format(output_format))
        return output_format

    def __coord_transform__(self, coord, source_crs):
        """
        Transforms the specified coordinates from the specified CRS to the configured target
        CRS and creates a point geometry.

        Args:
            coord (tuple): The coordinates to transform (x, y).
            source_crs (intorstr): The source CRS

        Returns:
            shapely.geometry.Point or shapely.geometry.Polygon: The transformed coordinates as
            Point.
        """
        log.debug('----- Transforming Coordinates: -----')
        log.debug('----- X/Y Coordinates: {0} -----'.format(coord))
        epsg = 'epsg:{0}'
        srid = Config.get('srid')
        log.debug('----- srid from config (to_srs): {0} -----'.format(srid))
        log.debug('----- srid from source (from_srs): {0} -----'.format(source_crs))
        rp = Reprojector()
        x, y = rp.transform(coord, from_srs=epsg.format(source_crs), to_srs=epsg.format(srid))
        log.debug('----- X/Y coordinates after transformation: ({0}, {1}) -----'.format(x, y))
        return Point(x, y)

    def __get_egrid_response__(self, records, params):
        """
        Creates a valid GetEGRID response from a list of real estate records.

        Args:
            records (list of pyramid_oereb.lib.records.real_estate.RealEstateRecord): List of real
                estate records.
            params (pyramid_oereb.views.webservice.Parameter): The parameter object.

        Returns:
            pyramid.response.Response: The `getegrid` response.
        """

        if len(records) == 0:
            return HTTPNoContent()
        output_format = self.__validate_format_param__(self._DEFAULT_FORMATS)
        real_estates = list()
        supported_languages = Config.get_language()
        for r in records:
            real_estate_type = Config.get_real_estate_type_by_data_code(getattr(r, 'type'))
            text = []
            for lang in real_estate_type.title:
                if lang in supported_languages:
                    text.append({
                        'Language': lang,
                        'Text': real_estate_type.title[lang]
                    })
            real_estate = {
                'egrid': getattr(r, 'egrid'),
                'number': getattr(r, 'number'),
                'identDN': getattr(r, 'identdn'),
                'type': {
                    'Code': real_estate_type.code,
                    'Text': text
                }
            }
            if params.with_geometry and output_format == 'json':
                real_estate.update({
                    'limit': Renderer.from_shapely(getattr(r, 'limit'))
                })
            elif params.with_geometry and output_format == 'xml':
                real_estate.update({
                    'limit': getattr(r, 'limit')
                })
            elif params.with_geometry and output_format not in ['json', 'xml']:
                raise HTTPInternalServerError('Format for GetEgrid not correct: {}'.format(output_format))
            real_estates.append(real_estate)
        egrid = {'GetEGRIDResponse': real_estates}
        if output_format == 'json':
            response = render_to_response('json', egrid, request=self._request)
            response.content_type = 'application/json; charset=UTF-8'
        else:
            response = render_to_response(
                'pyramid_oereb_getegrid_xml',
                (egrid, params),
                request=self._request
            )

        response.extras = OerebStats(service='GetEGRID', output_format=output_format)
        return response

    @staticmethod
    def __parse_en__(en, buffer_dist=None):
        """
        Parses the coordinates from the XY parameter and creates a point geometry.
        If a buffer distance is defined, a buffer with the specified distance will be applied.

        Args:
            en (str): EN parameter from the getegrid request.
            buffer_dist (float or None): Distance for the buffer applied to the
                point. If None, no buffer will be applied.

        Returns:
            shapely.geometry.Point or shapely.geometry.Polygon: The coordinates as
                Point or, if using a buffer, as Polygon.
        """
        coords = en.split(',')

        if len(coords) != 2:
            raise HTTPBadRequest(
                'The parameter EN has to be a comma-separated pair of coordinates.')

        x = float(coords[0])
        y = float(coords[1])
        p = Point(x, y)
        if buffer_dist:
            return p.buffer(buffer_dist)
        else:
            return p

    def __parse_gnss__(self, gnss):
        """
        Parses the coordinates from the GNSS parameter, transforms them to target CRS and
        creates a Point with a 1 meter buffer.

        Args:
            gnss (str): GNSS parameter from the getegrid request.

        Returns:
            shapely.geometry.Point or shapely.geometry.Polygon: The transformed coordinates as
            Point.
        """
        coords = gnss.split(',')

        if len(coords) != 2:
            raise HTTPBadRequest(
                'The parameter GNSS has to be a comma-separated pair of coordinates.')

        # Coordinates provided as "latitude,longitude"
        return self.__coord_transform__(coords, 4326).buffer(1.0)

    def __has_params__(self, needed):
        """
        Checks if the request contains all needed parameters.

        Args:
            needed (list of str): The parameters to check.

        Returns:
            bool: True if all needed parameters are available, false otherwise.
        """
        for p in needed:
            if p not in self._params:
                return False
        return True

    @staticmethod
    def __redirect_to_dynamic_client__(real_estate):
        """
        Returns a redirect to the configured dynamic client.

        Args:
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord):
                The found real estate.

        Returns:
            pyramid.httpexceptions.HTTPFound: The redirect response.
        """
        url = Config.get_extract_config().get('redirect')
        if url is None:
            raise HTTPInternalServerError('Missing configuration for redirect to dynamic client.')
        return HTTPFound(url.format(**vars(real_estate)))


class Parameter(object):
    def __init__(self, response_format, with_geometry=False, images=False, signed=False, identdn=None,
                 number=None, egrid=None, language=None, topics=None, extract_url=None, qr_code_ref=None):
        """
        Creates a new parameter instance.

        Args:
            response_format (str): The extract format.
            with_geometry (bool): Extract with/without geometry.
            images (bool): Extract with/without images.
            signed (bool): True for a signed extract.
            identdn (str): The IdentDN as real estate identifier.
            number (str): The parcel number as real estate identifier.
            egrid (str): The EGRID as real estate identifier.
            language (str): The requested language.
            topics (list of str): The list of requested topics.
            extract_url (str): The URL which was used to create the extract.
            qr_code_ref (str): The URL where the QR code can be downloaded from.
        """
        self.__format__ = response_format
        self.__with_geometry__ = with_geometry
        self.__images__ = images
        self.__signed__ = signed
        self.__identdn__ = identdn
        self.__number__ = number
        self.__egrid__ = egrid
        self.__language__ = language
        self.__topics__ = topics
        self.__extract_url__ = extract_url
        self.__qr_code_ref__ = qr_code_ref

    def set_identdn(self, identdn):
        """
        Updates the IdentDN.

        Args:
            identdn (str): The IdentDN as real estate identifier.
        """
        self.__identdn__ = identdn

    def set_number(self, number):
        """
        Updates the parcel number.

        Args:
            number (str): The parcel number as real estate identifier.
        """
        self.__number__ = number

    def set_egrid(self, egrid):
        """
        Updates the EGRID.

        Args:
            egrid (str): The EGRID as real estate identifier.
        """
        self.__egrid__ = egrid

    def set_language(self, language):
        """
        Updates the language.

        Args:
            language (str): The requested language.
        """
        self.__language__ = language

    def set_topics(self, topics):
        """
        Updates the requested topics.

        Args:
            topics (list of str): The list of requested topics.
        """
        self.__topics__ = topics

    def set_appurl(self, appurl):
        """
        Set the application base url

        Args:
            appurl (str): The baseurl of the application.
        """
        self.__appurl__ = appurl

    @property
    def format(self):
        """
        Returns:
            str: The requested format.
        """
        return self.__format__

    @property
    def with_geometry(self):
        """
        Returns:
            bool: Extract requested with geometries.
        """
        return self.__with_geometry__

    @property
    def images(self):
        """
        Returns:
            bool: Extract requested with images.
        """
        return self.__images__

    @property
    def signed(self):
        """
        Returns:
            bool: Signed extract requested.
        """
        return self.__signed__

    @property
    def identdn(self):
        """
        Returns:
            str: The requested IdentDN.
        """
        return self.__identdn__

    @property
    def number(self):
        """
        Returns:
            str: The requested number.
        """
        return self.__number__

    @property
    def egrid(self):
        """
        Returns:
            str: The requested EGRID.
        """
        return self.__egrid__

    @property
    def language(self):
        """
        Returns:
            str: The requested language.
        """
        return self.__language__

    @property
    def topics(self):
        """
        Returns:
            list of str: The requested topics.
        """
        return self.__topics__

    @property
    def extract_url(self):
        """

        Returns:
            str: The extract url
        """
        return self.__extract_url__

    @property
    def qr_code(self):
        """
        Returns:
            str: The QR code as binary encoded string.
        """

        return QRcode.create_qr_code(self.extract_url)

    @property
    def qr_code_ref(self):
        """
        Returns:
            str: The QR code as binary encoded string.
        """

        return self.__qr_code_ref__

    def skip_topic(self, theme_code):
        """
        Check if the topic should be skipped in extract.

        Args:
            theme_code (str): The PLR theme code.

        Returns:
            bool: True if the topic should be skipped.
        """
        if not self.topics or 'ALL' in self.topics:
            return False
        if theme_code in self.topics:
            return False
        if 'ALL_FEDERAL' in self.topics and theme_code in Config.get_all_federal():
            return False
        return True

    def __str__(self):
        return '<%s -- format: %s geometry: %s images: %s identdn: %s' \
                    ' number: %s egrid: %s language: %s topics: %s>' % (
                        self.__class__.__name__,
                        self.format, self.with_geometry, self.images, self.identdn,
                        self.number, self.egrid, self.language, self.topics)


class Logo(object):
    """
    Webservice to deliver logo images.

    Args:
        request (pyramid.request.Request or pyramid.testing.DummyRequest): The pyramid request instance.
    """
    def __init__(self, request):
        self._request = request

    def get_image(self):
        """
        Returns a response containing the binary logo image content using the configured "get_logo_method".

        Returns:
            pyramid.response.Response: Response containing the binary image content.
        """
        logo_key = self._request.matchdict.get('logo')
        logo_language = self._request.matchdict.get('language')
        if(logo_key.upper() == 'CONFEDERATION'):
            logo = Config.get_conferderation_logo()
        elif(logo_key.upper() == 'OEREB'):
            logo = Config.get_oereb_logo()
        elif(logo_key.upper() == 'CANTON'):
            logo = Config.get_canton_logo()
        elif(logo_key.upper() == 'MUNICIPALITY'):
            fosnr = self._request.params['fosnr']
            logo = Config.get_municipality_logo(fosnr)
        else:
            raise HTTPNotFound('This logo does not exist.')
        response = self._request.response
        response.status_int = 200
        response.body = logo.image_dict[logo_language].content
        response.content_type = logo.image_dict[logo_language].mimetype
        return response


class Symbol(object):
    """
    Webservice to deliver legend entry images.

    Args:
        request (pyramid.request.Request or pyramid.testing.DummyRequest): The pyramid request instance.
    """
    def __init__(self, request):
        self._request_ = request

    def get_image(self):
        """
        Returns a response containing the binary image content using the configured "get_symbol_method".

        Returns:
            pyramid.response.Response: Response containing the binary image content.
        """
        theme_code = self._request_.matchdict.get('theme_code')
        method = self.get_method(theme_code)
        theme_config = Config.get_theme_config_by_code(str(theme_code))
        if method:
            body, mimetype = method(dict(self._request_.params), theme_config)
            response = self._request_.response
            response.status_int = 200
            response.body = body
            response.content_type = mimetype
            return response
        log.error('"get_symbol_method" not found')
        raise HTTPNotFound()

    @staticmethod
    def get_method(theme_code):
        dnr = DottedNameResolver()
        theme_config = Config.get_theme_config_by_code(str(theme_code))
        return dnr.resolve(theme_config.get('hooks').get('get_symbol'))


class Sld(object):
    """
    Webservice to deliver a valid sld content to filter a av base layer to only one dedicated real estate.

    """

    def __init__(self, request):
        """
        Args:
            request (pyramid.request.Request or pyramid.testing.DummyRequest): The pyramid request instance.
        """
        self._request_ = request

    def get_sld(self):
        """
        Webservice which delivers an SLD file from parameter input. However
        this is a proxy pass through only. We use it to call the real method
        configured in the dedicated yaml file and hope that this method is
        accepting a pyramid.request.Request as input and is returning a
        pyramid.response.Response which encapsulates a well designed SLD.

        .. note:: The config path to define this hook method is:
            *pyramid_oereb.real_estate.visualisation.method*

        Returns:
             pyramid.response.Response: The response provided by the hooked method provided by the
                configuration

        Raises:
            pyramid.httpexceptions.HTTPInternalServerError: When the return
            value of the hooked method was not of type pyramid.response.Response
            pyramid.httpexceptions.HTTPNotFound: When the configured method was not found.
        """
        real_estate_config = Config.get_real_estate_config()
        params = {}
        for param in real_estate_config['visualisation']['url_params']:
            params.update({param: self._request_.params.get(param)})
        method = self.get_method()
        response = self._request_.response
        response.content_type = 'application/xml'
        if method:
            response.body = method(params, real_estate_config)
            return response
        log.error(u'method in path "{path}" not found'.format(
            path=real_estate_config['visualisation']['method'])
        )
        raise HTTPNotFound()

    @staticmethod
    def get_method():
        dnr = DottedNameResolver()
        visualisation_config = Config.get_real_estate_config().get('visualisation')
        method_path = visualisation_config.get('method')
        return dnr.resolve(method_path)


class QRcode(object):
    """
    Webservice to deliver a QR code image to recall the PDF extract of the defined real estate.

    """

    def __init__(self, request):
        """
        Args:
            request (pyramid.request.Request or pyramid.testing.DummyRequest): The pyramid request instance.
        """
        self._request_ = request

    def get_qr_code(self):

        extract_url = self._request_.params.get('extract_url')

        if extract_url is None:
            raise HTTPNoContent('No URL for QR Code generation was passed')
        extract_url = self.sanitize_url(extract_url)

        qr_code = self.create_qr_code(extract_url)
        response = self._request_.response
        response.status_int = 200
        response.body = qr_code
        response.content_type = 'image/png'  # buffered.mimetype
        return response

    @staticmethod
    def sanitize_url(url):
        # TODO: implement
        return url

    @staticmethod
    def create_qr_code(text):
        """
        Returns a binary image - the QR code.

        Args:
            url (text): The text which will be wrapped into the QR code.
        Returns:
            str: Binary image content as binary string.
        """
        # Create an object to the qrcode using the QRCode() function and store it in a
        # variable.
        qr = qrcode.QRCode()
        # Add data to the above QRcode uisng the add_data() function by passing some
        # random string as an argument.
        qr.add_data(text)
        # Get or build the QRcode using the make() function
        qr.make()
        # Convert the QRcode into an image using the make_image() function
        # store it in another variable.
        qr_img = qr.make_image()
        # Save the above image with some random name using the save() function
        # qr_img.save('myqrcode.png')
        buffered = io.BytesIO()
        qr_img.save(buffered, format="PNG")
        qr_code = buffered.getvalue()
        return qr_code



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.law_status import LawStatusRecord


class LawStatusBaseSource(Base):
    """
    Base class for law status values source.
    Attributes:
        records (list of pyramid_oereb.lib.records.law_status.LawStatusRecord): List of law status
        records.
    """
    _record_class_ = LawStatusRecord

    def read(self, params=None):
        """
        Every law status source has to implement a read method. This method must accept no parameters.
        Because it should deliver all items available.
        If you want adapt to your own source for real estate type labels, this is the point where to hook in.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.document_types import DocumentTypeRecord


class DocumentTypesBaseSource(Base):
    """
    Base class for document type values source.

    Attributes:
        records (list of pyramid_oereb.lib.records.document_types.DocumentTypeRecord): List of document
        type records.
    """
    _record_class_ = DocumentTypeRecord

    def read(self):
        """
        Every document type source has to implement a read method. This method must accept no parameters.
        Because it should deliver all items available.
        If you want adapt to your own source for document type labels, this is the point where to hook in.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-

from sqlalchemy.schema import CreateTable
from sqlalchemy.dialects import postgresql


def tables(base):
    return base.metadata.sorted_tables


def create_schema_sql(schema_name):
    """
    Args:
        schema_name (str): name of the schema

    Returns:
        str: query string to create the schema
    """
    return 'CREATE SCHEMA IF NOT EXISTS {0};'.format(schema_name)


def create_tables_sql(tables_to_create, if_not_exists=False):
    """
    Args:
        tables_to_create (list of sqlalchemy.schema.Table): The table objects from sqlalchemy.
        if_not_exists (bool): defaults to false. Determines if the table is created if it already exists.
            ... CREATE IF NOT EXISTS tables_to_create ...

    Returns:
        a string with the sql statement used to create the tables
    """
    sqls = []
    for table in tables_to_create:
        sqls.append(
            '{};\n'.format(
                str(CreateTable(table, if_not_exists=if_not_exists)
                    .compile(dialect=postgresql.dialect())).replace('DATETIME', 'timestamp')
            )
        )
    return ''.join(sqls)


def create_sql(schema_name, tables_to_create, if_not_exists=False):
    """
    Args:
        schema_name (str): name of the schema to be used.
        tables_to_create (list of sqlalchemy.schema.Table): The table objects from sqlalchemy.
        if_not_exists (bool): defaults to false. Determines if the table is created if it already exists.
            ... CREATE IF NOT EXISTS tables_to_create ...

    Returns:
        a string with the sql statement used to create the schema and the tables
    """
    sqls = [
        create_schema_sql(schema_name),
        create_tables_sql(tables_to_create, if_not_exists)
    ]
    return '\n'.join(sqls)



# -*- coding: utf-8 -*-
from geoalchemy2.elements import _SpatialElement
from geoalchemy2.shape import to_shape
from sqlalchemy.orm.exc import NoResultFound

from pyramid_oereb.core.sources import BaseDatabaseSource
from pyramid_oereb.core.sources.address import AddressBaseSource


class DatabaseSource(BaseDatabaseSource, AddressBaseSource):

    def read(self, params, street_name, zip_code, street_number):
        """
        The read method to access the standard database structure. It uses SQL-Alchemy for querying. It tries
        to find the items via passed arguments. If there was no result found it goes on with assigning an
        empty list as records instance attribute.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            street_name (unicode): The name of the street for the desired address.
            zip_code (int): The postal zipcode for the desired address.
            street_number (str): The house or so called street number of the desired address.
        """
        session = self._adapter_.get_session(self._key_)
        try:
            query = session.query(self._model_)
            results = [query.filter(
                self._model_.street_name == street_name
            ).filter(
                self._model_.zip_code == zip_code
            ).filter(
                self._model_.street_number == street_number
            ).one()]

            self.records = []
            for result in results:
                self.records.append(self._record_class_(
                    result.street_name,
                    result.zip_code,
                    result.street_number,
                    to_shape(result.geom) if isinstance(result.geom, _SpatialElement) else None
                ))

        except NoResultFound:
            self.records = []

        finally:
            session.close()



import datetime
import re
from mako.template import Template
from pyramid.path import AssetResolver

from pyramid_oereb import route_prefix
from pyramid_oereb.core.records.office import OfficeRecord


def get_symbol(params, theme_config):
    """
    This is a dummy method only to define what is expected by the real implementation.

    It is expected that this method delivers the actual binary content of the found image
    and the mime type fitting to the image content.

    Args:
        params (dict): The URL parameters which were handed over via request.
        theme_config (dict): The configuration of the theme how it was set up in the YAML.

    Returns:
        bytearray, str: The image content and the mimetype of image.
    """
    raise NotImplementedError('Method has to be implemented.')


def get_symbol_ref(request, record):
    """
    Returns the link to the symbol of the specified public law restriction.

    Args:
        request (pyramid.request.Request): The current request instance.
        record (pyramid_oereb.core.records.view_service.LegendEntryRecord): The legend entry record to get
            the symbol reference (link/url) for.

    Returns:
        uri: The link to the symbol for the specified public law restriction.
    """
    query = {
        'identifier': record.identifier
    }
    return request.route_url(
        '{0}/image/symbol'.format(route_prefix),
        theme_code=record.theme.code,
        extension=record.symbol.extension,
        _query=query
    )


def get_surveying_data_provider(real_estate):
    """

    Args:
        real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real estate for which the
            provider of the surveying data should be delivered.
    Returns:
        provider (pyramid_oereb.core.records.office.OfficeRecord): The provider who produced the used
            surveying data.
    """
    provider = OfficeRecord({u'de': u'This is only a dummy'})
    return provider


def get_surveying_data_update_date(real_estate):
    """
    Gets the date of the latest update of the used survey data data for the
    situation map. The method you find here is only matching the standard configuration. But you can provide
    your own one if your configuration is different. The only thing you need to take into account is that the
    input of this method is always and only a real estate record. And the output of this method must be a
    datetime.date object.

    Args:
        real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real
            estate for which the last update date of the base data should be indicated

    Returns:
        update_date (datetime.datetime): The date of the last update of the cadastral base data
    """

    update_date = datetime.datetime.now()

    return update_date


def produce_sld_content(params, real_estate_config):
    """
    This is the standard hook method to provide the sld content. Of course you can set it to another one. For
    instance to use another template. Or to use other parameters to provide the correctly constructed SLD.

    The SLD can be used as Styling parameter to the WMS call to produce a highlighted real estate.

    When you are replacing this method take care that it has to accept the parameters as defined here as input
    and must deliver a valid SLD. It is in your responsibility.

    Args:
        params (dict): The URL params which are passed.
        real_estate_config (dict): The configuration of the real estate how it was set up in the YAML.

    Returns:
        str: The rendered SLD (XML) as text.
    """
    template = Template(
        filename=AssetResolver('pyramid_oereb').resolve('core/views/templates/sld.xml').abspath(),
        input_encoding='utf-8',
        output_encoding='utf-8'
    )
    template_params = {
        'layer_name': re.sub(
            '<.*?>', '', real_estate_config['visualisation']['layer']['name'], flags=re.DOTALL
        ),
        'identifier': re.sub(
            '<.*?>', '', params['egrid'], flags=re.DOTALL
        ),
        'stroke_opacity': re.sub(
            '<.*?>', '', real_estate_config['visualisation']['style']['stroke_opacity'], flags=re.DOTALL
        ),
        'stroke_color': re.sub(
            '<.*?>', '', real_estate_config['visualisation']['style']['stroke_color'], flags=re.DOTALL
        ),
        'stroke_width': re.sub(
            '<.*?>', '', real_estate_config['visualisation']['style']['stroke_width'], flags=re.DOTALL
        )
    }
    return template.render(**template_params)


def plr_sort_within_themes(extract):
    """
    This is the standard hook method to sort a plr list (while respecting the theme order).
    This standard hook does no sorting, you can set your configuration to a different method if you need a
    specific sorting.

    Args:
        extract (pyramid_oereb.lib.records.extract.ExtractRecord): The unsorted extract

    Returns:
        pyramid_oereb.lib.records.extract.ExtractRecord: Returns the updated extract
    """
    return extract



# -*- coding: utf-8 -*-


class LawStatusRecord(object):
    """
    Law status record.

    Attributes:
        code (str or unicode): The code of the law status. It must be "inForce" or
            "changeWithPreEffect" or "changeWithoutPreEffect" every other value will
            raise an error.
        title (dict of unicode): The multilingual law status description.
    """

    def __init__(self, code, title):
        """
        Args:
            code (str or unicode): The code of the law status. It must be "inForce" or
                "changeWithPreEffect" or "changeWithoutPreEffect" every other value will
                raise an error.
            title (dict of unicode): The multilingual law status description.
        """
        self.code = code
        self.title = title

    def __str__(self):
        return '<{} -- code: {} title: {}>'.format(
            self.__class__.__name__, self.code, self.title)



from sqlalchemy import Column, ForeignKey
from sqlalchemy import LargeBinary, String, Integer, Date, Text
from sqlalchemy.ext.declarative import declarative_base
from geoalchemy2.types import Geometry as GeoAlchemyGeometry
from sqlalchemy.orm import relationship


class Models(object):

    def __init__(self, office, document, view_service,
                 legend_entry, public_law_restriction, geometry,
                 public_law_restriction_document,
                 localised_blob, localised_uri, multilingual_blob, multilingual_uri,
                 base, db_connection, schema_name):

        self.Office = office
        self.Document = document
        self.ViewService = view_service
        self.LegendEntry = legend_entry
        self.PublicLawRestriction = public_law_restriction
        self.Geometry = geometry
        self.PublicLawRestrictionDocument = public_law_restriction_document
        self.LocalisedBlob = localised_blob
        self.LocalisedUri = localised_uri
        self.MultilingualBlob = multilingual_blob
        self.MultilingualUri = multilingual_uri
        self.Base = base
        self.db_connection = db_connection
        self.schema_name = schema_name


def generic_models(base, schema_name, pk_type):
    """
    Factory to produce a set of generic standard models.

    Args:
        base (): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs

    Returns:
        list: First element is the Office model and second is the Document model.
    """

    class Office(base):
        """
        The bucket to fill in all the offices you need to reference from public law restriction, document,
        geometry.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually.
                        If you  don't like it - don't care about.
            t_ili_tid (str): TID from the transfer file.
            name (text): The name of the office. Interlis type: LocalisationCH_V1.MultilingualText.
            name_de (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            name_fr (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            name_it (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            name_rm (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            name_en (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            uid (str): The uid of this office from https
            line1 (str): The first address line for this office.
            line2 (str): The second address line for this office.
            street (str): The streets name of the offices address.
            number (str): The number on street.
            postal_code (int): The ZIP-code.
            city (str): The name of the city.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'amt'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        t_ili_tid = Column(String, nullable=True)
        name = Column('aname', Text, nullable=True)
        name_de = Column('aname_de', Text, nullable=True)
        name_fr = Column('aname_fr', Text, nullable=True)
        name_it = Column('aname_it', Text, nullable=True)
        name_rm = Column('aname_rm', Text, nullable=True)
        name_en = Column('aname_en', Text, nullable=True)
        uid = Column('auid', String(12), nullable=True)
        line1 = Column('zeile1', String, nullable=True)
        line2 = Column('zeile2', String, nullable=True)
        street = Column('strasse', String, nullable=True)
        number = Column('hausnr', String, nullable=True)
        postal_code = Column('plz', Integer, nullable=True)
        city = Column('ort', String, nullable=True)

    class Document(base):
        """
        THE DOCUMENT
        This represents the main document in the whole system.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually.
                        If you  don't like it - don't care about.
            t_ili_tid (str): TID from the transfer file.
            document_type (str): The document type. It must be "Rechtsvorschrift", "GesetzlicheGrundlage"
                or "Hinweis".
            title (text): The multilingual title of his document.
                        Interlis type: LocalisationCH_V1.MultilingualText.
            title_de (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            title_fr (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            title_it (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            title_rm (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            title_en (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            abbreviation (text): The multilingual shortened version of the documents title.
                                Interlis type: LocalisationCH_V1.MultilingualText.
            abbreviation_de (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            abbreviation_fr (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            abbreviation_it (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            abbreviation_rm (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            abbreviation_en (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            official_number (text): The multilingual official number which uniquely identifies this document.
                                    Interlis type: LocalisationCH_V1.MultilingualText.
            official_number_de (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            official_number_fr (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            official_number_it (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            official_number_rm (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            official_number_en (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            only_in_municipality (int): The fosnr (=id bfs) of the municipality. If this is None it is assumed
                the document is  related to the whole canton or even the confederation.
            index (int): An index used to sort the documents.
            law_status (str): The status switch if the document is legally approved or not.
            published_from (datetime.date): The date from when the document should be available for
                publishing in extracts. This  directly affects the behaviour of extract
                generation.
            published_until (datetime.date): The date until when the document should be available for
                publishing on extracts. This  directly affects the behaviour of extract
                generation.
            office_id (int): The foreign key to the office which is in charge for this document.
            responsible_office (pyramid_oereb.standard.models.railways_project_planning_zones.Office):
                The dedicated relation to the office instance from database.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'dokument'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        t_ili_tid = Column(String, nullable=True)
        document_type = Column('typ', String, nullable=False)
        title = Column('titel', Text, nullable=True)
        title_de = Column('titel_de', Text, nullable=True)
        title_fr = Column('titel_fr', Text, nullable=True)
        title_it = Column('titel_it', Text, nullable=True)
        title_rm = Column('titel_rm', Text, nullable=True)
        title_en = Column('titel_en', Text, nullable=True)
        abbreviation = Column('abkuerzung', Text, nullable=True)
        abbreviation_de = Column('abkuerzung_de', Text, nullable=True)
        abbreviation_fr = Column('abkuerzung_fr', Text, nullable=True)
        abbreviation_it = Column('abkuerzung_it', Text, nullable=True)
        abbreviation_rm = Column('abkuerzung_rm', Text, nullable=True)
        abbreviation_en = Column('abkuerzung_en', Text, nullable=True)
        official_number = Column('offiziellenr', Text, nullable=True)
        official_number_de = Column('offiziellenr_de', Text, nullable=True)
        official_number_fr = Column('offiziellenr_fr', Text, nullable=True)
        official_number_it = Column('offiziellenr_it', Text, nullable=True)
        official_number_rm = Column('offiziellenr_rm', Text, nullable=True)
        official_number_en = Column('offiziellenr_en', Text, nullable=True)
        only_in_municipality = Column('nuringemeinde', Integer, nullable=True)
        index = Column('auszugindex', Integer, nullable=False)
        law_status = Column('rechtsstatus', String, nullable=False)
        published_from = Column('publiziertab', Date, nullable=False)
        published_until = Column('publiziertbis', Date, nullable=True)
        office_id = Column(
            'zustaendigestelle',
            ForeignKey(Office.t_id),
            nullable=False
        )
        responsible_office = relationship(Office)

    return [Office, Document]


def model_factory(schema_name, pk_type, srid, db_connection):
    """
    Factory to produce a set of standard models.

    Args:
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.

    Returns:
        Models: the produced set of standard models

    """
    Base = declarative_base()

    Office, Document = generic_models(Base, schema_name, pk_type)

    class ViewService(Base):
        """
        A view service aka WM(T)S which can deliver a cartographic representation via web.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'darstellungsdienst'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)

    class LegendEntry(Base):
        """
        A class based legend system which is directly related to
        :class:`pyramid_oereb.standard.models.land_use_plans.ViewService`.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            symbol (str): An image with represents the legend entry. This can be png or svg. It is string
                but BaseCode64  encoded.
            legend_text (dict): Multilingual text to describe this legend entry.
            legend_text_de (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            legend_text_fr (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            legend_text_it (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            legend_text_rm (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.
            legend_text_en (text): Mapping of sources type: LocalisationCH_V1.MultilingualText.

            type_code (str): Type code of the public law restriction which is represented by this legend
                entry.
            type_code_list (str): List of all public law restrictions which are described through this
                legend  entry.
            theme (str): Statement to describe to which public law restriction this legend entry
                belongs.
            sub_theme (dict): Multilingual description for sub topics this legend entry might belonging to.
            view_service_id (str): The foreign key to the view service this legend entry is related to.
            view_service (pyramid_oereb.standard.models.land_use_plans.ViewService):
                The dedicated relation to the view service instance from database.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'legendeeintrag'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        symbol = Column(LargeBinary, nullable=False)
        legend_text = Column('legendetext', Text, nullable=True)
        legend_text_de = Column('legendetext_de', Text, nullable=True)
        legend_text_fr = Column('legendetext_fr', Text, nullable=True)
        legend_text_it = Column('legendetext_it', Text, nullable=True)
        legend_text_rm = Column('legendetext_rm', Text, nullable=True)
        legend_text_en = Column('legendetext_en', Text, nullable=True)
        type_code = Column('artcode', String(40), nullable=False)
        type_code_list = Column('artcodeliste', String, nullable=False)
        theme = Column('thema', String, nullable=False)
        sub_theme = Column('subthema', String, nullable=True)
        view_service_id = Column(
            'darstellungsdienst',
            ForeignKey(ViewService.t_id),
            nullable=False
        )
        view_service = relationship(ViewService, backref='legends')

    class PublicLawRestriction(Base):
        """
        The container where you can fill in all your public law restrictions to the topic.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            law_status (str): The status switch if the document is legally approved or not.
            published_from (datetime.date): The date when the document should be available for
                publishing on extracts. This  directly affects the behaviour of extract
                generation.
            published_until (datetime.date): The date starting from which the document should not be
                published anymore on extracts. This directly affects the behaviour of extract generation.
            view_service_id (int): The foreign key to the view service this public law restriction is
                related to.
            view_service (pyramid_oereb.standard.models.land_use_plans.ViewService):
                The dedicated relation to the view service instance from database.
            legend_entry_id (int): The foreign key to the legend entry this public law restriction is
                related to.
            legend_entry (pyramid_oereb.standard.models.airports_building_lines.LegendEntry):
                The dedicated relation to the legend entry instance from database.
            office_id (int): The foreign key to the office which is responsible to this public law
                restriction.
            responsible_office (pyramid_oereb.standard.models.land_use_plans.Office):
                The dedicated relation to the office instance from database.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'eigentumsbeschraenkung'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        law_status = Column('rechtsstatus', String, nullable=False)
        published_from = Column('publiziertab', Date, nullable=False)
        published_until = Column('publiziertbis', Date, nullable=True)
        view_service_id = Column(
            'darstellungsdienst',
            ForeignKey(ViewService.t_id),
            nullable=False
        )
        view_service = relationship(
            'ViewService',
            backref='public_law_restrictions'
        )
        legend_entry_id = Column(
            'legende',
            ForeignKey(LegendEntry.t_id),
            nullable=False
        )
        legend_entry = relationship(
            'LegendEntry',
            backref='public_law_restrictions')
        office_id = Column(
            'zustaendigestelle',
            ForeignKey(Office.t_id),
            nullable=False
        )
        responsible_office = relationship(Office)

    class Geometry(Base):
        """
        The dedicated model for all geometries in relation to their public law restriction.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            point (geoalchemy2.types.Geometry): Interlis type: GeometryCHLV95_V1.Coord2
            line (geoalchemy2.types.Geometry): Interlis type: GeometryCHLV95_V1.Line
            surface (geoalchemy2.types.Geometry): Interlis type: GeometryCHLV95_V1.Surface
            law_status (str): The status switch if the document is legally approved or not.
            published_from (datetime.date): The date when the geometry should be available for
                publishing on extracts. This directly affects the behaviour of extract
                generation.
            published_until (datetime.date): The date from when the geometry should not be available
                anymore for publishing on extracts. This directly affects the behaviour of extract
                generation.
            geo_metadata (str): A link to the metadata which this geometry is based on which delivers
                machine  readable response format (XML).
            public_law_restriction_id (str): The foreign key to the public law restriction this geometry
                is related to.
            public_law_restriction (pyramid_oereb.standard.models.land_use_plans
                .PublicLawRestriction): The dedicated relation to the public law restriction instance from
                database.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'geometrie'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        point = Column('punkt', GeoAlchemyGeometry('POINT', srid=srid), nullable=True)
        line = Column('linie', GeoAlchemyGeometry('LINESTRING', srid=srid), nullable=True)
        surface = Column('flaeche', GeoAlchemyGeometry('POLYGON', srid=srid), nullable=True)
        law_status = Column('rechtsstatus', String, nullable=False)
        published_from = Column('publiziertab', Date, nullable=False)
        published_until = Column('publiziertbis', Date, nullable=True)
        geo_metadata = Column('metadatengeobasisdaten', String, nullable=True)
        public_law_restriction_id = Column(
            'eigentumsbeschraenkung',
            ForeignKey(PublicLawRestriction.t_id),
            nullable=False
        )
        public_law_restriction = relationship(
            PublicLawRestriction,
            backref='geometries'
        )

    class MultilingualUri(Base):
        """
        Bucket to resolve entries related to OeREBKRM_V2_0.MultilingualUri.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            t_seq (int): Order of the structure elements.
            language (str): Interlis: InternationalCodes_V1.LanguageCode_ISO639_1.
            text (str): Interlis: URI.
            office_id (int): The foreign key to the office this multilingual uri
                is related to.
            document_id (int): The foreign key to the document this multilingual uri
                is related to.
            view_service_id (int): The foreign key to the view service this multilingual uri
                is related to.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'multilingualuri'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        t_seq = Column(Integer, nullable=True)
        office_id = Column(
            'amt_amtimweb',
            ForeignKey(Office.t_id),
            nullable=True
        )
        document_id = Column(
            'dokument_textimweb',
            ForeignKey(Document.t_id),
            nullable=True
        )
        view_service_id = Column(
            'darstellungsdienst_verweiswms',
            ForeignKey(ViewService.t_id),
            nullable=True
        )
        office = relationship(
            'Office',
            backref='multilingual_uri'
        )
        document = relationship(
            'Document',
            backref='multilingual_uri'
        )
        view_service = relationship(
            'ViewService',
            backref='multilingual_uri'
        )

    class MultilingualBlob(Base):
        """
        Bucket to resolve entries related to OeREBKRM_V2_0.MultilingualBlob.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            t_seq (int): Order of the structure elements.
            document_id (int): The foreign key to the document this multilingual uri
                is related to.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'multilingualblob'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        t_seq = Column(Integer, nullable=True)
        document_id = Column(
            'dokument_dokument',
            ForeignKey(Document.t_id),
            nullable=True
        )

    class LocalisedUri(Base):
        """
        Bucket to resolve entries related to OeREBKRM_V2_0.LocalisedUri.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            t_seq (int): Order of the structure elements.
            language (str): Interlis: InternationalCodes_V1.LanguageCode_ISO639_1.
            text (str): Interlis: URI.
            multilingualuri_id (int): The foreign key to the multilingual uri this localised uri
                is related to.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'localiseduri'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        t_seq = Column(Integer, nullable=True)
        language = Column('alanguage', String, nullable=True)
        text = Column('atext', String, nullable=False)
        multilingualuri_id = Column(
            'multilingualuri_localisedtext',
            ForeignKey(MultilingualUri.t_id),
            nullable=False
        )
        multilingualuri = relationship(
            'MultilingualUri',
            backref='localised_uri'
        )

    class LocalisedBlob(Base):
        """
        Bucket to resolve entries related to OeREBKRM_V2_0.LocalisedBlob.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            t_seq (int): Order of the structure elements.
            language (str): Interlis: InternationalCodes_V1.LanguageCode_ISO639_1.
            text (str): Interlis: URI.
            multilingualblob_id (int): The foreign key to the multilingual blob this localised blob
                is related to.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'localisedblob'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        language = Column('alanguage', String, nullable=True)
        blob = Column('ablob', LargeBinary, nullable=False)
        multilingualblob_id = Column(
            'multilingualblob_localisedblob',
            ForeignKey(MultilingualBlob.t_id),
            nullable=True
        )
        multilingualblob = relationship(
            'MultilingualBlob',
            backref='localised_uri'
        )

    class PublicLawRestrictionDocument(Base):
        """
        Meta bucket (join table) for the relationship between public law restrictions and documents.

        Attributes:
            t_id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            public_law_restriction_id (int): The foreign key to the public law restriction which has
                relation to  a document.
            document_id (int): The foreign key to the document which has relation to the public law
                restriction.
            plr (pyramid_oereb.standard.models.land_use_plans.PublicLawRestriction):
                The dedicated relation to the public law restriction instance from database.
            document (pyramid_oereb.standard.models.land_use_plans.DocumentBase):
                The dedicated relation to the document instance from database.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'hinweisvorschrift'
        t_id = Column(pk_type, primary_key=True, autoincrement=False)
        public_law_restriction_id = Column(
            'eigentumsbeschraenkung',
            ForeignKey(PublicLawRestriction.t_id),
            nullable=False
        )
        document_id = Column(
            'vorschrift',
            ForeignKey(Document.t_id),
            nullable=False
        )
        plr = relationship(
            PublicLawRestriction,
            backref='legal_provisions'
        )
        document = relationship(
            Document
        )

    return Models(
        Office, Document, ViewService,
        LegendEntry, PublicLawRestriction, Geometry, PublicLawRestrictionDocument,
        LocalisedBlob, LocalisedUri, MultilingualBlob, MultilingualUri,
        Base, db_connection, schema_name
    )


def model_factory_integer_pk(schema_name, geometry_type, srid, db_connection):
    """
    Args:
        schema_name (str): The name of the database schema where this models belong to.
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.
        db_connection (str): the db connection string

    Returns:
        Models: the produced set of standard models
    """
    if geometry_type is not None:
        geometry_type = None
    return model_factory(schema_name, Integer, srid, db_connection)


def model_factory_string_pk(schema_name, geometry_type, srid, db_connection):
    """
    Args:
        schema_name (str): The name of the database schema where this models belong to.
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.
        db_connection (str): the db connection string

    Returns:
        Models: the produced set of standard models
    """
    if geometry_type is not None:
        geometry_type = None
    return model_factory(schema_name, String, srid, db_connection)



# -*- coding: utf-8 -*-
from pyramid.path import DottedNameResolver


class DisclaimerReader(object):
    """
    The central reader for the disclaimer definitions. It is directly bound to a so called source
    which is defined by a pythonic dotted string to the class definition of this source.
    An instance of the passed source will be created on instantiation of this reader class by passing through
    the parameter kwargs.
    """

    def __init__(self, dotted_source_class_path, **params):
        """
        Args:
            dotted_source_class_path (str or
                pyramid_oereb.core.records.disclaimer.DisclaimerRecord):
                The path to the class which represents the source used by this reader. This
                class must exist and it must implement basic source behaviour of the
                :ref:`api-pyramid_oereb-core-sources-disclaimer-disclaimerbasesource`.
            (kwargs): kwargs, which are necessary as configuration parameter for the above by
                dotted name defined class.
        """
        source_class = DottedNameResolver().maybe_resolve(dotted_source_class_path)
        self._source_ = source_class(**params)

    def read(self):
        """
        The read method of this reader. There we invoke the read method of the bound source.

        Note:
            If you subclass this class your implementation needs to offer this method in the same
            signature. Means the parameters must be the same and the return must be a list of
            :ref:`api-pyramid_oereb-core-records-availability-availabilityrecord`. Otherwise the API like way
            the server works would be broken.

        Returns:
            list of pyramid_oereb.core.records.disclaimer.DisclaimerRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.
        """
        self._source_.read()
        return self._source_.records



# -*- coding: utf-8 -*-


class ThemeDocumentRecord(object):
    """Creates a new theme record.
    Attributes:
        theme_id (str): Foreign key relating to theme's id.
        document_id (str): Foreign key relating to document's id.
        article_numbers (list of str): Article numbers to refine statement of document to a theme.
    """

    def __init__(self, theme_id, document_id, article_numbers):
        """
        Args:
            theme_id (str): Foreign key relating to theme's id.
            document_id (str): Foreign key relating to document's id.
            article_numbers (list of str): Article numbers to refine statement of document to a theme.
        """

        self.theme_id = theme_id
        self.document_id = document_id
        self.article_numbers = article_numbers

    def __str__(self):
        return '<{} -- theme_id: {} document_id: {}>'.format(
            self.__class__.__name__, self.theme_id, self.document_id)



# -*- coding: utf-8 -*-

import logging

log = logging.getLogger(__name__)


def eliminate_duplicated_document_records(main_document_records, plr_document_records):
    """ Filtering of document records that are associated to a plr.

    Document records associated to a plr are eliminated if a record associated to a theme exists
    for the same document. Document records associated to a theme have priority.
    Records are considered to handle the same document if:
    - indices are equal, and
    - document_type codes are equal, and
    - official_numbers correspond

    Correct data is expected.
    """

    # basic rules (one or the other source does not provide any document records)
    if main_document_records is None and len(plr_document_records) > 0:
        return plr_document_records
    if main_document_records is not None and len(plr_document_records) == 0:
        return main_document_records

    # list which indicates duplicated documents
    plr_document_is_duplicated_list = [False] * len(plr_document_records)

    # document per document comparison
    for doc1 in main_document_records:
        for index2, doc2 in enumerate(plr_document_records):
            if plr_document_is_duplicated_list[index2] is False:
                # comparison of indices
                if doc1.index != doc2.index:
                    continue
                # comparison of document_type
                if doc1.document_type.code != doc2.document_type.code:
                    continue
                # comparison of number
                # - Note: official number is NOT mandatory
                # - possibility of not corresponding languages
                docs_have_same_number = False
                if doc1.official_number is not None and doc2.official_number is not None:
                    for key, value in doc1.official_number.items():
                        if doc2.official_number.get(key):
                            if doc2.official_number.get(key) == value:
                                docs_have_same_number = True
                                break
                    if docs_have_same_number:
                        plr_document_is_duplicated_list[index2] = True

    unique_document_indexes = [i for i, val in enumerate(plr_document_is_duplicated_list) if val is False]
    unique_document_records = [plr_document_records[i] for i in unique_document_indexes]

    # logging message if document record was removed
    if len(unique_document_records) != len(plr_document_records):
        dupl_doc_indexes = [i for i, val in enumerate(plr_document_is_duplicated_list) if val is True]
        for i in dupl_doc_indexes:
            log.info(
                '''PLR document record removed from extract as it is already provided by the main doc records
                (title: {title}, number: {number}, index: {index})'''.format(
                    title=plr_document_records[i].title,
                    number=plr_document_records[i].official_number,
                    index=plr_document_records[i].index
                )
            )

    return main_document_records + unique_document_records



# -*- coding: utf-8 -*-
from mako.lookup import TemplateLookup
from pyramid.path import AssetResolver

from pyramid.response import Response

from pyramid_oereb.core.renderer import Base
from pyramid_oereb.core.views.webservice import Parameter
from mako import exceptions


class Renderer(Base):

    def __init__(self, info):
        """
        Creates a new XML renderer instance for versions rendering.

        Args:
            info (pyramid.interfaces.IRendererInfo): Info object.
        """
        a = AssetResolver('pyramid_oereb')
        resolver = a.resolve('core/renderer/getegrid/templates/xml')
        geom_resolver = a.resolve('core/renderer/extract/templates/xml')
        self.template_dirs = [resolver.abspath(), geom_resolver.abspath()]
        self._gml_id = 0
        super(Renderer, self).__init__(info)

    def __call__(self, value, system):
        """
        Returns the XML encoded versions response according to the specification.

        Args:
            value (tuple): A tuple containing the real estates and the params
                dictionary.
            system (dict): The available system properties.

        Returns:
            str: The XML encoded versions response.
        """
        self._request = self.get_request(system)
        response = self.get_response(system)
        if isinstance(response, Response) and response.content_type == response.default_content_type:
            response.content_type = 'application/xml'

        self._params_ = value[1]
        if not isinstance(self._params_, Parameter):
            raise TypeError('Missing parameter definition; Expected {0}, got {1} instead'.format(
                Parameter,
                self._params_.__class__
            ))

        templates = TemplateLookup(
            directories=self.template_dirs,
            output_encoding='utf-8',
            input_encoding='utf-8'
        )
        template = templates.get_template('getegrid.xml')
        try:
            content = template.render(**{
                'data': value[0],
                'params': self._params_,
                'sort_by_localized_text': self.sort_by_localized_text,
                'localized': self.get_localized_text,
                'multilingual': self.get_multilingual_text,
                'get_localized_image': self.get_localized_image,
                'request': self._request
            })
            return content
        except Exception:
            response.content_type = 'text/html'
            return exceptions.html_error_template().render()



# -*- coding: utf-8 -*-

import logging

from pyramid_oereb.core.adapter import DatabaseAdapter
from pyramid_oereb.core.config import Config
from pyramid.config import Configurator

__version__ = '1.0.1'


log = logging.getLogger(__name__)
route_prefix = None
# initially instantiate database adapter for global session handling
database_adapter = DatabaseAdapter()


def main(global_config, **settings):
    """
    This function returns a Pyramid WSGI application. This is necessary for development of
    your plugin. So you can run it local with the paster server and in a IDE like PyCharm. It
    is intended to leave this section as is and do configuration in the includeme section only.
    Push additional configuration in this section means it will not be used by the production
    environment at all!
    """
    config = Configurator(settings=settings)
    config.include('pyramid_oereb', route_prefix='oereb')
    config.scan()
    return config.make_wsgi_app()


def includeme(config):
    """
    By including this in your pyramid web app you can easily provide a running OEREB Server

    Args:
        config (Configurator): The pyramid apps config object
    """

    global route_prefix

    # Set route prefix
    route_prefix = config.route_prefix

    # Get settings
    settings = config.get_settings()

    # Load configuration file
    cfg_file = settings.get('pyramid_oereb.cfg.file', None)
    cfg_c2ctemplate_file = settings.get('pyramid_oereb.cfg.c2ctemplate.file', None)
    cfg_section = settings.get('pyramid_oereb.cfg.section', None)
    Config.init(
        cfg_file or cfg_c2ctemplate_file,
        cfg_section, cfg_file is None,
        init_data=True
    )
    Config.update_settings(settings)

    settings.update({
        'pyramid_oereb': Config.get_config()
    })

    config.add_renderer('pyramid_oereb_extract_json', 'pyramid_oereb.core.renderer.extract.json_.Renderer')
    config.add_renderer('pyramid_oereb_extract_xml', 'pyramid_oereb.core.renderer.extract.xml_.Renderer')
    config.add_renderer('pyramid_oereb_extract_print', Config.get('print').get('renderer'))
    config.add_renderer('pyramid_oereb_versions_xml', 'pyramid_oereb.core.renderer.versions.xml_.Renderer')
    config.add_renderer('pyramid_oereb_capabilities_xml',
                        'pyramid_oereb.core.renderer.capabilities.xml_.Renderer')
    config.add_renderer('pyramid_oereb_getegrid_xml', 'pyramid_oereb.core.renderer.getegrid.xml_.Renderer')

    config.include('pyramid_oereb.core.routes')



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.data_integration import DataIntegrationRecord


class DataIntegrationBaseSource(Base):
    """
    Base class for data integration sources.

    Attributes:
        records (list of pyramid_oereb.core.records.data_integration.DataIntegrationRecord): List
            of availability records.
    """
    _record_class_ = DataIntegrationRecord

    def read(self):
        """
        Every availability source has to implement a read method. This method must accept no
        parameters. Because it should deliver all items available.
        If you want to adapt to your own source for data integration, this is the point where
        to hook in.

        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
import warnings


class DocumentTypeRecord(object):

    """
    Represents a document type entry with the code and it's display name.

    Args:
        code (str of unicode): The code for the document type.
        title (dict of unicode): The label title for the document type (multilingual).
    """
    def __init__(self, code, title):
        if not isinstance(code, str):
            warnings.warn('Type of "code" should be "str"')
        if not isinstance(title, dict):
            warnings.warn('Type of "title" should be "dict"')

        self.code = code
        self.title = title

    def __str__(self):
        return '<{} -- code: {} title: {}>'.format(
            self.__class__.__name__, self.code, self.title)



# -*- coding: utf-8 -*-
import logging
from pyramid.path import DottedNameResolver

from shapely.geometry import box
from timeit import default_timer as timer

from pyramid_oereb.core.config import Config
from pyramid_oereb.core.records.extract import ExtractRecord
from pyramid_oereb.core.records.image import ImageRecord
from pyramid_oereb.core.records.plr import PlrRecord, EmptyPlrRecord

log = logging.getLogger(__name__)


class ExtractReader(object):
    """
    The class which generates *the extract* as a record
    (:ref:`api-pyramid_oereb-core-records-extract-extractrecord`). This is the point where all necessary
    and extract related components are bound together.

    Attributes:
        extract (pyramid_oereb.lib.records.extract.ExtractRecord or None): The extract as a record
            representation. On initialisation this is None. It will be set by calling the read method of the
            instance.
    """

    def __init__(self, plr_sources, plr_cadastre_authority):
        """
        Args:
            plr_sources (list of pyramid_oereb.lib.sources.plr.PlrBaseSource): The list of PLR source
                instances which the achieved extract should be about.
            plr_cadastre_authority (pyramid_oereb.lib.records.office.OfficeRecord): The authority responsible
                for the PLR cadastre.
        """
        self.extract = None
        self._plr_sources_ = plr_sources
        self._plr_cadastre_authority_ = plr_cadastre_authority
        self.law_status = Config.get_law_status_codes()

    @property
    def plr_cadastre_authority(self):
        """
        Returns:
            pyramid_oereb.lib.records.office.OfficeRecord: The authority responsible for the PLR
            cadastre.
        """
        return self._plr_cadastre_authority_

    def read(self, params, real_estate, municipality):
        """
        This method finally creates the extract.

        Note:
            If you subclass this class your implementation needs to offer this method in the same
            signature. Means the parameters must be the same and the return must be a
            :ref:`api-pyramid_oereb-core-records-extract-extractrecord`. Otherwise the API like way the server
            works would be broken.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real
                estate for which the report should be generated
            municipality (pyramid_oereb.lib.records.municipality.MunicipalityRecord): The municipality
                record.

        Returns:
            pyramid_oereb.lib.records.extract.ExtractRecord:
                The extract record containing all gathered data.
        """
        log.debug("read() start")

        bbox = Config.get_bbox(real_estate.limit)
        bbox = box(bbox[0], bbox[1], bbox[2], bbox[3])

        concerned_themes = list()
        not_concerned_themes = list()
        themes_without_data = list()

        if municipality.published:

            for plr_source in self._plr_sources_:
                if not params.skip_topic(plr_source.info.get('code')):
                    plr_source.read(params, real_estate, bbox)

                    real_estate.public_law_restrictions.extend(plr_source.records)

            for plr in real_estate.public_law_restrictions:

                # Filter topics due to topics parameter
                if not params.skip_topic(plr.theme.code):
                    if isinstance(plr, PlrRecord):
                        if plr.theme.code not in [theme.code for theme in concerned_themes]:
                            concerned_themes.append(plr.theme)
                    elif isinstance(plr, EmptyPlrRecord):
                        if plr.has_data:
                            not_concerned_themes.append(plr.theme)
                        else:
                            themes_without_data.append(plr.theme)

        else:
            for plr_source in self._plr_sources_:
                themes_without_data.append(Config.get_theme_by_code_sub_code(plr_source.info.get('code')))

        # sort plr according to theme, sub-theme and law-status
        start_time = timer()
        log.debug("sort plrs by theme and law status start")
        real_estate.public_law_restrictions.sort(key=lambda element: (
            self._sort_plr_theme(element), self._sort_plr_law_status(element)
        ))
        end_time = timer()
        log.debug(f"DONE with sort plrs by theme and law status, time spent: {end_time-start_time} seconds")

        # Load base data form configuration
        resolver = DottedNameResolver()
        date_method_string = Config.get('extract').get('base_data').get('methods').get('date')
        date_method = resolver.resolve(date_method_string)
        update_date_os = date_method(real_estate)
        general_information = Config.get_general_information()

        oereb_logo = Config.get_oereb_logo()
        confederation_logo = Config.get_conferderation_logo()
        canton_logo = Config.get_canton_logo()
        municipality_logo = Config.get_municipality_logo(municipality.fosnr)
        qr_code_image = ImageRecord(params.qr_code)

        self.extract = ExtractRecord(
            real_estate,
            oereb_logo,
            confederation_logo,
            canton_logo,
            municipality_logo,
            self.plr_cadastre_authority,
            update_date_os,
            concerned_theme=concerned_themes,
            not_concerned_theme=not_concerned_themes,
            theme_without_data=themes_without_data,
            general_information=general_information,
            qr_code=qr_code_image,
            qr_code_ref=params.qr_code_ref
        )

        log.debug("read() done")
        return self.extract

    def _sort_plr_law_status(self, plr_element):
        """
        This method generates the sorting key for plr_elements according to their law_status code.
        The value is generated from the index the plr_element.law_status.code has in the law_status
        list. The law_status list corresponds to the law status taken from the DB

        If the argument is not a PlrRecord or its law_status.code is not contained in the law_status list,
        the method will return the length of the law_status list so it can be sorted at the end of the list.

        Args:
            plr_element (PlrRecord or EmptyPlrRecord) a plr record element.

        Returns:
            int: Value which can be used to sort the record depending on its law_status.code.

        """
        if (isinstance(plr_element, PlrRecord) and plr_element.law_status.code in self.law_status):
            return self.law_status.index(plr_element.law_status.code)
        else:
            return len(self.law_status)

    @staticmethod
    def _sort_plr_theme(plr_element):
        """
        This method generates a sorting key to sort PLRs in to themes and sub-themes.
        The value is generated using the extract_index given for each theme.
        Currently it is assumed that the extract_index for a sub-theme is in accord to
        its theme so that the order will be correct.

        If the plr_element is not a PlrRecord 10 000 will be returned so that it is
        added at the end of the list

        Args:
            plr_element (PlrRecord or EmptyPlrRecord) a plr record element.

        Returns:
            int: Value which can be used to sort the record depending on its theme/sub-theme.
        """
        if (isinstance(plr_element, PlrRecord)):
            index = plr_element.sub_theme.extract_index \
                if (plr_element.sub_theme is not None) \
                else plr_element.theme.extract_index
            return index
        return 10000



# -*- coding: utf-8 -*-


class DataIntegrationRecord(object):

    def __init__(self, date, checksum=None, theme=None, office=None,
                 theme_identifier=None, office_identifier=None):
        """
        The record to check if data is available for municipality.

        Args:
            date (datetime.date): The date when data for the theme was delivered.
            theme (pyramid_oereb.core.records.ThemeRecord or None): The theme code
                which this element belongs to.
            office (pyramid_oereb.core.records.office.OfficeRecord or None): The office
                this element belongs to.
            checksum (str or None): A checksum to persist the data state which is in the
                db. It is thought to be a helper field to check if import is necessary.
            theme_identifier (str|int|None): The identifier which can be used to match the
                corresponding theme.
            office_identifier (str|int|None): The identifier which can be used to match the
                corresponding office.
        """
        if theme is None and theme_identifier is None:
            raise AttributeError('Either theme or theme_identifier has to be defined!')
        if office is None and office_identifier is None:
            raise AttributeError('Either office or  office_identifier has to be defined!')
        self.date = date
        self.theme = theme
        self.office = office
        self.checksum = checksum
        self.theme_identifier = theme_identifier
        self.office_identifier = office_identifier



# -*- coding: utf-8 -*-
import warnings
import logging
import requests

from pyramid_oereb.core.records.image import ImageRecord
from pyramid_oereb.core.url import add_url_params, parse_url
from pyramid_oereb.core.url import uri_validator
from shapely.geometry.point import Point
from pyramid_oereb.core import get_multilingual_element

log = logging.getLogger(__name__)


class LegendEntryRecord(object):
    """
    Represents a legend entry with it's text as well as it's image.

    Attributes:
        symbol (pyramid_oereb.lib.records.image.ImageRecord): The binary content of the legend symbol.
        legend_text (dict of unicode): The multilingual description text for the legend entry.
        type_code (unicode): The class of the legend entry corresponding to the plrs classes.
        type_code_list (unicode): An URL to the type code list.
        theme (pyramid_oereb.lib.records.theme.ThemeRecord): The theme to which the legend entry
            belongs.
        view_service_id (int): The id to the connected view service. This is very important to be able to
            solve bug https://github.com/openoereb/pyramid_oereb/issues/521
        sub_theme (pyramid_oereb.lib.records.theme.ThemeRecord): The optional sub theme to which the
            legend entry belongs.
        identifier (str): The identifier of the legend entry which might be used for linking to
                other elements.
    """

    def __init__(self, symbol, legend_text, type_code, type_code_list, theme,
                 view_service_id=None, sub_theme=None, identifier=None):
        """
        Args:
            symbol (pyramid_oereb.lib.records.image.ImageRecord): The binary content of the legend symbol.
            legend_text (dict of unicode): The multilingual description text for the legend entry.
            type_code (unicode): The class of the legend entry corresponding to the plrs classes.
            type_code_list (unicode): An URL to the type code list.
            theme (pyramid_oereb.lib.records.theme.ThemeRecord): The theme to which the legend entry
                belongs.
            view_service_id (int): The id to the connected view service. This is very important to be able to
                solve bug https://github.com/openoereb/pyramid_oereb/issues/521
            sub_theme (pyramid_oereb.lib.records.theme.ThemeRecord): The optional sub theme to which the
                legend entry belongs.
            identifier (str): The identifier of the legend entry which might be used for linking to
                    other elements.
        """
        if not isinstance(legend_text, dict):
            warnings.warn('Type of "legend_text" should be "dict"')

        self.symbol = symbol
        self.legend_text = legend_text
        self.type_code = type_code
        self.type_code_list = type_code_list
        self.theme = theme
        self.view_service_id = view_service_id
        self.sub_theme = sub_theme
        self.identifier = identifier

    def __str__(self):
        return '<{} -- symbol: {} legend_text: {} type_code: {} type_code_list: {}'\
                    ' theme: {} sub_theme: {} view_service_id: {} identifier: {}'\
                    .format(self.__class__.__name__, self.symbol, self.legend_text,
                            self.type_code, self.type_code_list, self.theme, self.sub_theme,
                            self.view_service_id, self.identifier)


class ViewServiceRecord(object):
    """
    A view service contains a valid WMS URL with a defined set of layers.

    Attributes:
        image (dict): multilingual dictionary containing the binary image
            (pyramid_oereb.core.records.image.ImageRecord) downloaded from WMS link for the
            requested (if any) or default language. Empty for an extract without images
    """

    def __init__(self, reference_wms, layer_index, layer_opacity, default_language,
                 srid, proxies=None, legends=None):
        """

        Args:
            reference_wms (dict): Multilingual dict with URLs to the actual service (WMS)
            layer_index (int): Layer index. Value from -1000 to +1000.
            layer_opacity (float): Opacity of layer. Value from 0.0 to 1.0.
            default_language (str): The default language which should be used for url selection.
            srid (int): The SRID which is used for the WMS.
            proxies (dict or None): The proxies which may be used
            legends (list of LegendEntry or None): A list of all relevant legend entries.
        """
        self.reference_wms = reference_wms
        self.image = dict()  # multilingual dict with binary map images resulting from calling the wms link

        self.layer_index = self.sanitize_layer_index(layer_index)
        self.layer_opacity = self.sanitize_layer_opacity(layer_opacity)

        self.min = None
        self.max = None
        self.calculate_ns()
        self.check_min_max_attributes(self.min, 'min', self.max, 'max')
        self.default_language = default_language
        self.srid = srid
        self.proxies = proxies

        if legends is None:
            self.legends = []
        else:
            for legend in legends:
                assert isinstance(legend.symbol, ImageRecord)
            self.legends = legends

    @staticmethod
    def sanitize_layer_index(layer_index):
        """
        Checks the validity of the layer index
        Arg:
           layer_index (int): index of the layer

        Returns:
            int: the layer index if it is an int berween -1000 and 1000.

        Raises:
            AttributeError if the index is out of bounds
        """
        if layer_index and not isinstance(layer_index, int):
            warnings.warn('Type of "layer_index" should be "int"')
        if layer_index < -1000 or layer_index > 1000:
            error_msg = "layer_index should be >= -1000 and <= 1000, " \
                        "was: {layer_index}".format(layer_index=layer_index)
            log.error(error_msg)
            raise AttributeError(error_msg)
        return layer_index

    @staticmethod
    def sanitize_layer_opacity(layer_opacity):
        """
        Checks the validity of the layer opacity
        Args:
            layer_opacity (float): the opacity used to draw the layer.

        Returns:
            float: the layer opacity if it is a float between 0 and 1

        Raises:
            AttributeError if the opacity is out of bounds.
        """
        if layer_opacity and not isinstance(layer_opacity, float):
            warnings.warn('Type of "layer_opacity" should be "float"')
        if layer_opacity < 0.0 or layer_opacity > 1.0:
            error_msg = "layer_opacity should be >= 0.0 and <= 1.0, " \
                        "was: {layer_opacity}".format(layer_opacity=layer_opacity)
            log.error(error_msg)
            raise AttributeError(error_msg)
        return layer_opacity

    @staticmethod
    def check_min_max_attributes(min_point, min_name, max_point, max_name):
        """
        Checks the validity of a min and max point:
            - type check
            - min_point < max_point
        Args:
            min_point (shapely.geometry.point.Point): a point geometry
            min_name (): the name of the point
            max_point (shapely.geometry.point.Point): a point geometry
            max_name ():the name of the point

        Raises:
            AttributeError if one of the check fails
        """
        if min_point is None and max_point is None:
            return
        if min_point is None or max_point is None:
            error_msg = 'Both {min_name} and {max_name} have to be defined'.format(min_name=min_name,
                                                                                   max_name=max_name)
            raise AttributeError(error_msg)
        if not isinstance(min_point, Point):
            raise AttributeError('Type of "{min_name}" should be "shapely.geometry.point.Point"'
                                 .format(min_name=min_name))
        if not isinstance(max_point, Point):
            raise AttributeError('Type of "{max_name}" should be "shapely.geometry.point.Point"'
                                 .format(max_name=max_name))
        if min_point.x > max_point.x or min_point.y > max_point.y:
            error_msg = 'Some value of {min_name} are larger than {max_name}'.format(min_name=min_name,
                                                                                     max_name=max_name)
            raise AttributeError(error_msg)

    def get_full_wms_url(self, language, map_width, map_height, bbox):
        """
        Returns the WMS URL to get the image.

        Args:
            language (string): which language of the reference WMS should be used.
            map_width (int): Width of map in pixels.
            map_height (int): Height of map in pixels.
            bbox (list of float): Bounding box which defines the maps view.

        Returns:
            str: The url used to query the WMS server.
        """

        if language not in self.reference_wms:
            msg = f"No WMS reference found for the requested language ({language}), using default language"
            log.info(msg)
            language = self.default_language

        self.reference_wms[language] = add_url_params(
            get_multilingual_element(self.reference_wms, language), {
                "BBOX": ",".join([str(e) for e in bbox]),
                "SRS": 'EPSG:{0}'.format(self.srid),
                "WIDTH": map_width,
                "HEIGHT": map_height
            }
        )
        self.calculate_ns()

        return self.reference_wms

    def download_wms_content(self, language):
        """
        Downloads the image found behind the URL stored in the instance attribute "reference_wms"
        for the requested language

        Args:
            language (string): the language for which the image should be downloaded

        Raises:
            LookupError: Raised if the response is not code 200 or content-type
                doesn't contains type "image".
            AttributeError: Raised if the URL itself isn't valid at all.
        """
        main_msg = "Image for WMS couldn't be retrieved."

        if language not in self.reference_wms:
            msg = f"No WMS reference found for the requested language ({language}), using default language"
            log.info(msg)
            language = self.default_language

        wms = self.reference_wms.get(language)

        if uri_validator(wms):
            log.debug(f"Downloading image, url: {wms}")
            try:
                response = requests.get(wms, proxies=self.proxies)
            except Exception as ex:
                dedicated_msg = f"An image could not be downloaded. URL was: {wms}, error was {ex}"
                log.error(dedicated_msg)
                raise LookupError(dedicated_msg)

            content_type = response.headers.get('content-type', '')
            if response.status_code == 200 and content_type.find('image') > -1:
                self.image[language] = ImageRecord(response.content)
            else:
                dedicated_msg = f"The image could not be downloaded. URL was: {wms}, " \
                    f"Response was {response.content.decode('utf-8')}"
                log.error(main_msg)
                log.error(dedicated_msg)
                raise LookupError(dedicated_msg)
        else:
            dedicated_msg = f"URL seems to be not valid. URL was: {wms}"
            log.error(main_msg)
            log.error(dedicated_msg)
            raise AttributeError(dedicated_msg)

    def calculate_ns(self):
        self.min, self.max = self.get_bbox_from_url(self.reference_wms[list(self.reference_wms.keys())[0]])

    @staticmethod
    def get_bbox_from_url(wms_url):
        """
        Parses wms url for BBOX parameter an returns these points as suitable values for ViewServiceRecord.
        Args:
            wms_url (str): wms url which includes a BBOX parameter to parse.

        Returns:
            set of two shapely.geometry.point.Point: min and max coordinates of bounding box.
        """
        _, params = parse_url(wms_url)
        bbox = params.get('BBOX')
        if bbox is None or len(bbox[0].split(',')) != 4:
            return None, None
        points = bbox[0].split(',')
        return Point(float(points[0]), float(points[1])), Point(float(points[2]), float(points[3]))



# -*- coding: utf-8 -*-

from pyramid.path import DottedNameResolver

from pyramid_oereb import Config
from pyramid_oereb.core.records.real_estate import RealEstateRecord
from pyramid_oereb.core.records.view_service import ViewServiceRecord


class RealEstateReader(object):
    """
    The central reader for real estates. It is directly bound to a so called source
    which is defined by a pythonic dotted string to the class definition of this source.
    An instance of the passed source will be created on instantiation of this reader class by passing through
    the parameter kwargs.
    """

    def __init__(self, dotted_source_class_path, **params):
        """
        Args:
            dotted_source_class_path (str or pyramid_oereb.lib.sources.real_estate.RealEstateBaseSource): The
                path to the class which represents the source used by this reader. This class must exist and
                it must implement basic source behaviour of the
                :ref:`api-pyramid_oereb-core-sources-real_estate-realestatebasesource`.
            (kwargs): kwargs, which are necessary as configuration parameter for the above by
                dotted name defined class.
        """
        source_class = DottedNameResolver().resolve(dotted_source_class_path)
        self._source_ = source_class(**params)

    def read(self, params, nb_ident=None, number=None, egrid=None, geometry=None):
        """
        The central read accessor method to get all desired records from configured source.

        Note:
            If you subclass this class your implementation needs to offer this method in the same
            signature. Means the parameters must be the same and the return must be a list of
            :ref:`api-pyramid_oereb-core-records-real_estate-realestaterecord`. Otherwise the API like way the
            server works would be broken.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            nb_ident (int or None): The identification number of the desired real estate. This
                parameter is directly related to the number parameter and both must be set!
                Combination will deliver only one result or crashes.
            number (str or None): The number of parcel or also known real estate. This parameter
                is directly related to the nb_ident parameter and both must be set!
                Combination will deliver only one result or crashes.
            egrid (str or None): The unique identifier of the desired real estate. This will deliver
                only one result or crashes.
            geometry (str or None): A geometry as WKT string which is used to obtain intersected real
                estates. This may deliver several results.

        Returns:
            list of pyramid_oereb.core.records.real_estate.RealEstateRecord:
                The list of all found records filtered by the passed criteria.
        """
        plan_for_land_register_config = Config.get_plan_for_land_register_config()

        real_estate_view_service = ViewServiceRecord(
            plan_for_land_register_config.get('reference_wms'),
            plan_for_land_register_config.get('layer_index'),
            plan_for_land_register_config.get('layer_opacity'),
            Config.get('default_language'),
            Config.get('srid'),
            Config.get('proxies')
        )

        plan_for_land_register_main_page_config = Config.get_plan_for_land_register_main_page_config()
        real_estate_main_page_view_service = ViewServiceRecord(
            plan_for_land_register_main_page_config.get('reference_wms'),
            plan_for_land_register_main_page_config.get('layer_index'),
            plan_for_land_register_main_page_config.get('layer_opacity'),
            Config.get('default_language'),
            Config.get('srid'),
            Config.get('proxies')
        )

        self._source_.read(params, nb_ident=nb_ident, number=number, egrid=egrid, geometry=geometry)
        for r in self._source_.records:
            if isinstance(r, RealEstateRecord):
                r.set_view_service(real_estate_view_service)
                r.set_main_page_view_service(real_estate_main_page_view_service)
        return self._source_.records



# -*- coding: utf-8 -*-
"""
This is package provides the minimum requirements on the classes which can be used as a source.
"""
import time
import logging
from pyramid.config import ConfigurationError
from pyramid.path import DottedNameResolver

log = logging.getLogger(__name__)


class Base(object):
    """
    The basic source class. This is not meant to be used directly as a source at runtime. But more as a basic
    class for inherit in special designed classes.

    Attributes:
        records (list): The list which will be filled up with records in the process.
    """
    records = list()


class BaseDatabaseSource(Base):
    """
    The plug for addresses which uses a database as source.

    Attributes:
        TIMEOUT (int): Seconds to wait until database should be ready
    """

    TIMEOUT = 60

    def __init__(self, **kwargs):
        """
        Keyword Args:
            db_connection (str): A rfc1738 conform database connection string in the form of:
                ``<driver_name>://<username>:<password>@<database_host>:<port>/<database_name>``
            model (str): A valid dotted name string which leads to an importable representation of
                sqlalchemy.ext.declarative.DeclarativeMeta or the real class itself.
        """
        from pyramid_oereb import database_adapter

        if database_adapter:
            self._adapter_ = database_adapter
        else:
            raise ConfigurationError('Adapter for database must be defined if you use database sources.')
        if kwargs.get('db_connection'):
            self._key_ = kwargs.get('db_connection')
        else:
            raise ConfigurationError('"db_connection" for source has to be defined in used yaml '
                                     'configuration file')
        if kwargs.get('model'):
            self._model_ = DottedNameResolver().maybe_resolve(kwargs.get('model'))
        else:
            raise ConfigurationError('"model" for source has to be defined in used yaml configuration file')
        # check if database is available and wait until it might be or raise error
        timeout_target = time.time() + self.TIMEOUT
        db_healthy = False
        current_wait_position = 0
        sleep_period = 1.0
        while time.time() < timeout_target:
            db_healthy = self.health_check()
            if not db_healthy:
                current_wait_position += int(sleep_period)
                # 1.0 sets it to a second
                log.info('Waiting for the database {} more seconds ({})'.format(
                    self.TIMEOUT - current_wait_position,
                    self._key_
                ))
                time.sleep(sleep_period)
            else:
                break
        if not db_healthy:
            raise ConfigurationError('Database was not reached until timeout of {} seconds ({})'.format(
                self.TIMEOUT,
                self._key_
            ))

    def get_session(self):
        return self._adapter_.get_session(self._key_)

    def health_check(self):
        session = self._adapter_.get_session(self._key_)
        try:
            session.execute('SELECT 1')
            return True
        except Exception:
            return False



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.real_estate import RealEstateRecord


class RealEstateBaseSource(Base):
    """
    Base class for real estate sources.

    Attributes:
        records (list of pyramid_oereb.lib.records.real_estate.RealEstateRecord): List of real estate records.
    """
    _record_class_ = RealEstateRecord

    def read(self, params, nb_ident=None, number=None, egrid=None, geometry=None):
        """
        Every real estate source has to implement a read method. This method must accept the four key word
        parameters. If you want adapt to your own source for real estates, this is the point where to hook in.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            nb_ident (int or None): The identification number of the desired real estate. This
                parameter is directly related to the number parameter and both must be set!
                Combination must deliver only one result or must raise an error.
            number (str or None): The number of parcel or also known real estate. This parameter
                is directly related to the nb_ident parameter and both must be set!
                Combination must deliver only one result or must raise an error.
            egrid (str or None): The unique identifier of the desired real estate. This must deliver only one
                result or must raise an error.
            geometry (str): A geometry as WKT string which is used to obtain intersected real
                estates. This may deliver several results.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.availability import AvailabilityRecord


class AvailabilityBaseSource(Base):
    """
    Base class for availability sources.

    Attributes:
        records (list of pyramid_oereb.core.records.availability.AvailabilityRecord): List
            of availability records.
    """
    _record_class_ = AvailabilityRecord

    def read(self):
        """
        Every availability source has to implement a read method. This method must accept no
        parameters. Because it should deliver all items available.
        If you want to adapt to your own source for availability, this is the point where to hook in.

        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.theme import ThemeRecord


class ThemeBaseSource(Base):
    """
    Base class for theme sources.

    Attributes:
        records (list of pyramid_oereb.lib.records.theme.ThemeRecord): List of theme records.
    """
    _record_class_ = ThemeRecord

    def read(self):
        """
        Every theme source has to implement a read method. This method must accept no parameters. Because
        it should deliver all items available.
        If you want adapt to your own source for themes, this is the point where to hook in.
        """
        pass  # pragma: no cover
